# -*- coding: iso-8859-1 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

'''
This module provides widgets L{EditableTreeWidget} L{TreeListWidget} and L{ObservableListWidget} for L{EditableTree}, L{ObservableSortedDictionary of EditableTree} and L{ObservableList} model objects.
These widgets register callback methods to update when the model changes.
They provide user interaction to modify the underlying model (drag&drop, contextual menu...)

@author: Dominique Geffroy
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''
__docformat__ = "epytext en"

import os
from StringIO import StringIO
from soma.qt_gui.qt_backend.QtGui import QTreeWidget, QTreeWidgetItem, QListWidget, QListWidgetItem, QPixmap, QDrag, QMenu, QPainter, QPen, QCursor, QSizePolicy, QIcon, qApp, QKeyEvent, QApplication
from soma.qt_gui.qt_backend.QtCore import Qt, QEvent, QMimeData, QObject, QPoint, QRect, QSize, QTimer, SIGNAL
import copy
from soma.notification import ObservableList, EditableTree
from soma.minf.api import defaultReducer, createMinfWriter, iterateMinf, minfFormat
from soma.wip.application.api import findIconFile
from soma.qt4gui.api import defaultIconSize

#----------------------------------------------------------------------------
class EditableTreeWidget(QTreeWidget):
  """A widget to represent a tree represented by an EditableTree object.
  The tree can be modifiable by manipulating items in the widget :
    - add new item
    - copy an item in a branch
    - move an item in a branch
    - delete an item

  It is possible to copy items from one tree widget to another.

  This component is created with a EditableTree object which is the model.

  The graphic component has a reference on the model and methods to control events.
  It registers a callback method on the model notifier.
  On drag&drop of an item in another, the model is updated and notify all its listeners,
  so the view is updated via updateContent method.

  The widgets treats some events :
    - drag&drop events : to copy or move items
    - key events : del to remove an item, shift to set move mode
    - context menu request event: show a popup menu to create new item
    - rename event on modifiable items.

  All items icon are resized to have the same size.

  Inner classes :
    - Item
    - Branch
    - Leaf

  @type MARGIN: int
  @cvar MARGIN: margin width in pixel around items

  @type model: EditableTree
  @ivar model: The tree which this widget is the representation
  @type controller: EditableTreeController
  @ivar controller: the controller is called to proceed model's changes on events.
  @type draggedItems: list of EditableTree.Item
  @ivar draggedItems: list of currently dragged items. Items can come from current tree or from another tree (so it is a copy of the other tree's item). 
  @type draggedItemsParent: list of EditableTree.Item
  @ivar draggedItemsParent: list of currently dragged items' parents (if they are in current tree). Usefull for delete and move functionalities.
  @type dropOn: EditableTree or Item
  @ivar dropOn: draggedItem is currently on this container
  @type dropAfter: Item
  @ivar dropAfter: draggedItem is currently after this item
  @type dropBefore: Item
  @ivar dropBefore: draggedItem is currently before this item
  @type popupMenu: QPopupMenu
  @ivar popupMenu: contextual menu shown on mouse right click
  @type contextMenuTarget: EdtiableTree.Item or EditableTree
  @ivar contextMenuTarget: item on which a context menu is opened
  """
  # margin in pixel around the items
  # it is used for the highlighting of the dropzones
  MARGIN=3

  def __init__(self, treeModel, parent=None, iconSize=defaultIconSize):
    """
    @type treeModel: EditableTree
    @param treeModel: the tree which current widget represents.
    @type parent:QWidget
    @param parent: parent of this widget
    @type iconSize: couple of integers or None
    @param iconSize: force items icon resizing.
    """
    QTreeWidget.__init__( self, parent)
    self.setColumnCount(1)
    self.setColumnWidth(0, self.width())
    self.setRootIsDecorated(True)
    #self.setItemMargin(self.MARGIN)
    #self.setSortingEnabled(False) # disable sort
    self.setSizePolicy( QSizePolicy( QSizePolicy.Preferred, QSizePolicy.Preferred ) )
    self.setIconSize(QSize(*iconSize))
    self.iconDim=iconSize
    # enable several items selection using ctrl and shift keys.
    self.setSelectionMode(QTreeWidget.ExtendedSelection)
    self.setAcceptDrops(True)
    # draggedItems attribute stores the items which are currently dragged
    self.draggedItems=[]
    self.draggedItemParent=[] # and their parents in current tree
    self.dragStartPosition=0
    # current drop zones (used in drawContentsOffset to highlight current dropzone)
    self.dropOn=None # dragged item will be dropped in this item's children
    self.dropAfter=None # dragged item will be dropped after this item (sibling)
    self.dropBefore=None # dragged item will be dropped before this item
    # Popup Menu
    self.popupMenu = QMenu(self)
    self.popupMenu.addAction( "New",  self.menuNewItemEvent )
    self.contextMenuTarget=None
    self.setContextMenuPolicy(Qt.CustomContextMenu)
    # keep a reference to the model for control event and register a listener to be aware of changes
    self.setModel(treeModel)
    
    self.connect(self, SIGNAL("itemChanged( QTreeWidgetItem *, int )"), self.itemRenamed)
    self.connect(self, SIGNAL( 'customContextMenuRequested ( const QPoint & )'), self.openContextMenu)

  def setModel(self, m):
    """Construct the children of the tree reading the model given.
    A listener is added to the model, so when the model change, the method self.updateContents
    is automatically called.
    """
    self.clear()
    self.model=m
    self.controller=EditableTreeController(m)
    if m!=None:
      self.setWindowTitle(m.name)
      self.setHeaderLabel(m.name)
      self.setAcceptDrops(m.modifiable)
      self.model.addListener(self.updateContent)
      self.model.onAttributeChange("name", self.updateName)
      # create child items with data in the tree model
      lastChild=None
      for item in self.model.values():
        if item.isLeaf(): # append item to keep the same order as in the model
          lastChild=EditableTreeWidget.Leaf(self, item, lastChild, self.iconDim)
        else: 
          lastChild=EditableTreeWidget.Branch(self, item, lastChild, self.iconDim)
  
  def itemRenamed(self, item, col):
    if item is not None and hasattr(item, "okRename"):
      item.okRename(col)
    
  def getChild(self, n):
    """Return the nth child item
    If n<=0 return the first child, if n>=nb children, return the last child
    @rtype: EditableTreeWidget.Item
    """
    if n<=0:
      n=0
    else:
      nbItems=self.topLevelItemCount()
      if n>=nbItems:
        n=nbItems-1
    return self.topLevelItem(n)

  def getLastChild(self):
    """Return the last child item.
    @rtype: EditableTreeWidget.Item
    """
    return self.topLevelItem(self.topLevelItemCount()-1)

  def toContentsPoint(self, point):
    """Translates coordinates in the frame in coordinates in the content of the list (without header).
    @rtype: QPoint"""
    #viewportPoint=self.contentsToViewport(point) to get the item at position we need the position in the entire list not in the visible content (viewport)
    return point # - QPoint(0, self.header().height())
  # headerItem().sizeHint().height() ?

  # already implemented in QTreeWidget
  #def selectedItems(self):
    #"""
    #Gets items that are currently selected in the listview (as we are in extended selection mode).
    
    #@rtype: list of EditableTreeWidget.Item
    #@return: items currently selected
    #"""
    #items=[]
    #it = QListViewItemIterator(self, QListViewItemIterator.Selected)
    #while it.current() :
        #items.append( it.current() )
        #it+=1
    #return items
    
  #------ Drag&Drop ------
  def mousePressEvent(self, event):
    if (event.button() == Qt.LeftButton):
      self.dragStartPosition = QPoint(event.pos())
    QTreeWidget.mousePressEvent(self, event)
    
  def mouseMoveEvent(self, event):
    """
    The QDrag object is shown during the drag.
    This method is called when the mouse moves, this can be the beginning of a drag if the left button is clicked and the distance between the current position and the dragStartPosition is sufficient.
    draggedItems attribute is set with selected item's model.
    It constructs a minf (xml) representation of drag objects which will be provided to the target drop zone.
    """
    if (not (event.buttons() & Qt.LeftButton)):
      return
    if ((event.pos() - self.dragStartPosition).manhattanLength()
          < QApplication.startDragDistance()):
      return
    
    items=self.selectedItems()
    # keep a reference to the current dragged item
    d=None
    if items != []:
      self.draggedItems=[]
      self.draggedItemsParent=[]
      for item in items:
        self.draggedItems.append(item.model)
        self.draggedItemsParent.append(item.container().model)
      #create a string containing the minf representation of items
      itemMinfBuf=StringIO()
      # find minf format to use for item's class
      firstItem=items[0].model
      reducer = defaultReducer( firstItem ) # supposing that all items needs the same reducer
      itemMinf=""
      if reducer != None:
        writer = createMinfWriter( itemMinfBuf, "XML", reducer )
        for item in items:
          writer.write( item.model )
        itemMinf=itemMinfBuf.getvalue()
        writer.close()
      #else: itemMinf=item.name
      # create a QTextDrag object with mime type text/xml containing the minf string
      d=QDrag(self)
      icon = findIconFile( firstItem.icon )
      if icon is not None: # adding an icon which will be visible on drag move, it will be the first item's icon
        d.setPixmap(QPixmap(icon))
      mimeData = QMimeData()
      mimeData.setText(itemMinf)
      d.setMimeData(mimeData);
      dropAction = d.exec_();
    #QTreeWidget.mouseMoveEvent(self, event)
      
  def dragEnterEvent(self, event):
    """This method is called when a drag enter in the widget. The source of the drag can be this widget or antoher.
    The event contains the encoding of the drag object.
    If the event's source is the current window, the event is always accepted.
    Else, try to decode text event as minf representation of an item and set self.draggedItem.
    """
    # if source of event is this widget, draggedItem is already set.
    if event.source()==self:
      event.accept()
    else: #must decode the event. If it is minf format, read it and initialize draggedItem
      self.draggedItems=[]
      self.draggedItemsParent=[]
      if event.mimeData().hasText():
        textEvent=event.mimeData().text()
        textEventBuf=StringIO(textEvent)
        # check if it is the expected minf format
        format, reduction=minfFormat(textEventBuf)
        if format=="XML":
          textEventBuf.seek(0)
          for value in iterateMinf( textEventBuf ):
            if isinstance(value, EditableTree.Item):
              event.accept()
              self.draggedItems.append(value)
            else: event.ignore()
          textEventBuf.close()
        else: event.ignore()
      else: event.ignore()

  def dragLeaveEvent(self, event):
    """The drag leave the widget, there's no more dropzones to highlight."""
    self.dropOn=None
    self.dropAfter=None
    self.dropBefore=None
    self.viewport().update()

  def dragMoveEvent(self, event):
    """When the drag moves, the current dropzone can change.
    According to cursor position (on, before, after an item),
    attributes that store current dropzone are updated (dropBefore, dropAfter, dropOn).
    Method update is called in order to refresh painting with dropzone highlighting.
    """
    # draggedItem must be set, otherwise it isn't a correct drag
    dropzonesChanged=False
    if self.draggedItems==[]:
      event.ignore(self.contentsRect())
      dropzonesChanged=self.setDropzones(None, None, None)
    else:
      cursorPos=self.toContentsPoint(event.pos())
      currentItem=self.itemAt(cursorPos) #item under the mouse cursor
      if currentItem!=None: #cursor on an item
        currentItemContainer=currentItem.container()
        rect = self.visualItemRect(currentItem)
        if currentItem.model not in self.draggedItems: #the item under the mouse isn't one of dragged items
          # if the cursor is in the top margin of item, insert before item
          if (cursorPos.y() < rect.top()+self.MARGIN):
            if currentItemContainer.acceptDrops(): # container must accept drop
              # in theory, it is possible to pass to accept method a rectangle for which the event is accepted,
              # in order to speed up the process event but this doesn't work :
              # when using this hint, dropzones highlighting is not always updated...
              #event.accept(QRect(rect.left(), rect.top(), rect.width(), self.MARGIN))
              event.accept()
              dropzonesChanged=self.setDropzones(currentItemContainer, currentItem, None)
            else:
              #event.ignore(QRect(rect.left(), rect.top(), rect.width(), self.MARGIN))
              event.ignore()
              dropzonesChanged=self.setDropzones(None, None, None)
          #if the cursor is in the bottom margin of the item -> drop after this item
          elif (cursorPos.y()>rect.bottom()-self.MARGIN):
            if currentItemContainer.acceptDrops():
              #event.accept(QRect(rect.left(), rect.bottom()-self.MARGIN, rect.width(), self.MARGIN))
              event.accept()
              dropzonesChanged=self.setDropzones(currentItemContainer, None, currentItem)
            else:
              #event.ignore(QRect(rect.left(), rect.bottom()-self.MARGIN, rect.width(), self.MARGIN))
              event.ignore()
              dropzonesChanged=self.setDropzones(None, None, None)
          # else if the cursor is on an item which accept drop
          elif currentItem.acceptDrops():
            #event.accept(QRect(rect.left(), rect.top()+self.MARGIN, rect.width(), rect.height()-2*self.MARGIN))
            event.accept()
            dropzonesChanged=self.setDropzones(currentItem, None, None)
          else:
            #event.ignore(QRect(rect.left(), rect.top()+self.MARGIN, rect.width(), rect.height()-2*self.MARGIN))
            event.ignore()
            dropzonesChanged=self.setDropzones(None, None, None)
        else: # item dragged on itself : ignore the event
          #event.ignore(rect)
          event.ignore()
          dropzonesChanged=self.setDropzones(None, None, None)
      elif self.acceptDrops(): #if the cursor is not on an item, the dropzone is the listview (at the end of the list)
        event.accept()
        dropzonesChanged=self.setDropzones(self, None, self.getLastChild())
      else:
        event.ignore()
        dropzonesChanged=self.setDropzones(None, None, None)
      if dropzonesChanged:
        self.viewport().update()

  def setDropzones(self, dropOn, dropBefore, dropAfter):
    """Sets dropOn, dropBefore and dropAfter attributes with parameter values.
    Return true if the dropzones have changed."""
    change=(self.dropOn != dropOn) or (self.dropBefore != dropBefore) or (self.dropAfter != dropAfter)
    if change:
      self.dropOn=dropOn
      self.dropBefore=dropBefore
      self.dropAfter=dropAfter
    return change

  def dropEvent( self, event):
    """When an item is dropped in another, a deep copy of the item is added in the dropzone item"""
    # current dropzone is stored in dropOn
    if self.dropOn!=None:
      dropAfterModel=None
      dropBeforeModel=None
      if self.dropAfter!=None: dropAfterModel=self.dropAfter.model
      if self.dropBefore!=None: dropBeforeModel=self.dropBefore.model
      self.controller.drop(event.source(), self, self.draggedItems, self.draggedItemsParent, self.dropOn.model, dropAfterModel, dropBeforeModel, ((event.keyboardModifiers() & Qt.ControlModifier) == Qt.ControlModifier) )
    self.draggedItems = []
    self.draggedItemsParent=[]
    self.dropOn=None
    self.dropAfter=None
    self.dropBefore=None
    self.viewport().update()

  def deleteDraggedItems(self):
    """Called when the item have been moved in another tree widget (drop+moveMode)"""
    i=0
    for draggedItem in self.draggedItems:
      self.controller.delete(draggedItem, self.draggedItemsParent[i])
      i+=1
    self.draggedItems = []
    self.draggedItemsParent=[]

  #------ Key events ------
    
  def keyPressEvent(self, event):
    """
    This method is called when a key is pressed during a few seconds or when a key is released
    If the user keep the key pressed, the event occurs however;
    same as pressing several times the key. It doesn't seems to be parametrable...
    If the key pressed is delete, the selected items are deleted.
    """
    if (event.key() == Qt.Key_Delete):
      # delete current selected item
      items = self.selectedItems()
      for item in items:
        self.controller.delete(item.model, item.container().model)
    else: event.ignore() # the event could be handled by a parent component

  #------ context menu events ------
  def openContextMenu(self, point):
    """On right click on the mouse, a context menu opens.
    With this menu, it is possible to create a new branch.
    """
    self.contextMenuTarget=None
    cursorPos=self.toContentsPoint(point)
    currentItem=self.itemAt(cursorPos) #item under the mouse cursor
    accept=False
    if currentItem is not None:
      if currentItem.model.modifiable and not currentItem.model.isLeaf():
        self.contextMenuTarget=currentItem.model
        accept=True
    else:
      if self.model is not None and self.model.modifiable:
        self.contextMenuTarget=self.model
        accept=True
    #items=self.selectedItems()
    #if len(items)==1:
      #if items[0].model.modifiable:
        #event.accept()
      #else: event.ignore()
    #elif self.model!=None and self.model.modifiable:
      #event.accept()
    #else: event.ignore()
    if accept:
      self.popupMenu.exec_(QCursor.pos())

  def menuNewItemEvent(self):
    """
    Called when user selects New in context menu in order to create a new item in the tree.
    """
    #items=self.selectedItems()
    #if len(items) == 1:
      #target=items[0].model
    #else: target=self.model
    #self.controller.newItem(target)
    self.controller.newItem(self.contextMenuTarget)

  #------ Update on model changes ------
  def updateContent(self, action, items, position=None):
    """This method is called when the model notifies a change :
    The action has been done at position, with items.
    The view should update its content to reflect the changes.
    """
    if action==ObservableList.INSERT_ACTION:
      self.insert(items, position)
    elif action==ObservableList.REMOVE_ACTION:
      self.remove(items, position)
    elif action==ObservableList.MODIFY_ACTION:
      # some model items have been modified,
      # widget items are replaced by new items with new model items
      # from position, all items must be replaced with new value
      # ->remove and then insert new
      i=position+1
      for modelItem in items:
        self.takeTopLevelItem(i)
        i+=1
      self.insert(items, position)
    #else: print action, "unknown action"

  def updateName(self, newName):
    self.setWindowTitle(newName)
    self.setHeaderLabel(newName)

  def insert(self, items, position=None):
    """Insertion of items at position in the model
    -> create view items for all items and insert them at position in self.
    Inserted item becomes the selected item."""
    #insert at position = insert before item at position = insert after item at position-1
    for item in items:
      if item.isLeaf(): # append item to keep the same order as in the model
        itemBefore=EditableTreeWidget.Leaf(None, item, None, self.iconDim)
      else: itemBefore=EditableTreeWidget.Branch(None, item, None, self.iconDim)
      self.insertTopLevelItem(position, itemBefore)
      if item.unamed and self.hasFocus(): 
        self.editItem(itemBefore, 0)
        item.unamed=False
    #print "current item : ", self.currentItem()
    #print "set current item : ", itemBefore, itemBefore.text(0)
    self.setCurrentItem(itemBefore, 0)
    itemBefore.setSelected(True)

  def remove(self, items, position):
    """Removes items in the list from position.
    If the list is empty, one item is removed.
    If position is undefined, removes items whose model is in the list"""
    if position is None:
      position=0
    if len(items)==0:
      self.takeTopLevelItem(position)
    else:
      for modelItem in items:
        found=False
        i=0
        while not found and i<self.topLevelItemCount():
          item=self.topLevelItem(i)
          if item.model is modelItem:
            self.takeTopLevelItem(i)
            found=True
          i+=1

  #------ Refresh painting ------
  def paintEvent(self, event):
    """This method is called to paint the tree.
    To refresh the view, call self.update.
    It is redefined in order to highlight drop zone during the drag of an item
    Three types of dropzone are defined :
      - before an item (cursor is in the top of the item's rectangle) -> draw a line on top
      - after an item (cursor is in the bottom of the item's rectangle) -> draw a line on bottom
      - on an item -> draw a rectangle around item
    """
    QTreeWidget.paintEvent(self, event)
    painter=QPainter()
    painter.begin(self.viewport())
    if self.dropBefore!=None: # draw a line in the top of the item
      rect=self.visualItemRect(self.dropBefore)
      # offset of the item = depth * offset in relation to parent
      #offset=self.indentation()
      painter.setBrush(Qt.black)
      # the line has the same offset as the item
      painter.drawRect(rect.left(), rect.top(), rect.width(), self.MARGIN)
    elif self.dropAfter!=None: # draw a line in the bottom of the item and its visible content
      rect=self.visualItemRect(self.dropAfter) # this rect only contains the item, not its content
      # inc the height to include item's content
      #rect.setHeight( (min( self.dropAfter.totalHeight(), self.viewport().height() - rect.y() ) ) ) # stay in the viewport to keep the line visible
      #offset=self.indentation()
      painter.setBrush(Qt.black)
      #print "draw drop after", self.dropAfter.getText(), rect.left(), rect.top(), rect.width(), rect.height()
      painter.drawRect(rect.left(), rect.bottom()-self.MARGIN+1, rect.width(), self.MARGIN)
    elif self.dropOn!=None: #draw the rectangle containing the item
      if self.dropOn!=self:
        rect=self.visualItemRect(self.dropOn)
      else: #ajout dans le listview alors qu'il est vide
        rect=QRect(0, 0, self.columnWidth(0), self.MARGIN)
      #print "draw drop on", self.dropOn.getText(), rect.left(), rect.top(), rect.width(), rect.height()
      painter.setPen(QPen(Qt.black, self.MARGIN, Qt.SolidLine))
      painter.drawRect(rect)
    painter.end()
    
    
  #----------------------------------------------------------------------------
  class Item(QTreeWidgetItem):
    """Item is the base class for elements of EditableTreeWidget. """
    def __init__( self, parent, treeItemModel, after=None, iconSize=defaultIconSize):
      """
      @type parent: tree or branch
      @param parent: container of the item
      @type treeItemModel: EditableTree.Item
      @param treeItemModel: model which this widget is the representation
      @type after: item
      @param after: the item after which current item must be added in the parent
      """
      QTreeWidgetItem.__init__( self, parent, [treeItemModel.name] )
      if treeItemModel.icon:
        iconPath = findIconFile( treeItemModel.icon )
        if iconPath is not None:
          self.setIcon(0, QIcon(iconPath))
      if not treeItemModel.copyEnabled:
        self.setFlags(self.flags() & ~Qt.ItemIsDragEnabled)
      if not treeItemModel.modifiable:
        self.setFlags(self.flags() & ~Qt.ItemIsDropEnabled)
      # rename is enabled only if the item is modifiable
      if treeItemModel.modifiable:
        self.setFlags(self.flags() | Qt.ItemIsEditable)
      treeItemModel.onAttributeChange("name", self.updateName)
      treeItemModel.onAttributeChange("visible", self.updateVisibiliy)
      self.setHidden(not treeItemModel.visible)
      self.setDisabled(not treeItemModel.enabled)
      self.model=treeItemModel

    def acceptDrops(self):
      return ((self.flags() & Qt.ItemIsDropEnabled) == Qt.ItemIsDropEnabled)

    def getText(self):
      return self.text(0)

    def container(self):
      """Return the parent item or the listview that contains the element if it is a top level element
      @rtype: EditableTreeWidget.Branch or EditableTreeWidget"""
      parent = self.parent()
      if parent is None:
        parent = self.treeWidget()
      return parent

    def okRename(self, col):
      """It is the called associated to the signal QTreeWidget.itemChanged(item, col). It is called when user renames the item.
      The name must be changed in the model."""
      if getattr(self, "model", None) is not None:
        newText=unicode(self.getText())
        if self.model.name != newText:
          if self.model.name == self.model.tooltip:
            self.model.tooltip=newText
          self.model.name=newText
    
    def updateName(self, newName):
      """
      Called when the model notifies that its name attribute value has changed.
      """
      self.setText(0, newName)

    def updateVisibiliy(self, newValue):
      """
      Called when the model notifies that its visible attribute value has changed.
      """
      #pass
      self.setHidden(not newValue)
      #self.setText(0, self.getText()+"-hid")
      #print "change visibility to ", newValue, " for ", self.getText()
  #----------------------------------------------------------------------------
  class Branch( Item ):
    """Item that represents a tree branch.
    It can have children."""
    def __init__( self, parent, treeItemModel, after=None, iconSize=defaultIconSize):
      EditableTreeWidget.Item.__init__(self, parent, treeItemModel, after, iconSize)
      self.iconDim=iconSize
      treeItemModel.addListener(self.updateContent)
      #create child items
      lastChild=None
      for item in treeItemModel.values():
        if item.isLeaf():
          lastChild=EditableTreeWidget.Leaf(self, item, lastChild, iconSize)
        else: lastChild=EditableTreeWidget.Branch(self, item, lastChild, iconSize)

    def getChild(self, n):
      """Return the nth child item
      If n<=0 return the first child, if n>=nb children, return the last child
      @rtype: EditableTreeWidget.Item
      """
      if n<0:
        n=0
      elif n>=self.childCount():
        n=self.childCount()-1
      return self.child(n)

    #------ Update on model changes ------
    def updateContent(self, action, items, position=None):
      """This method is called when the model notifies a change :
      The action has been done at position, with items.
      The view should update its content to reflect the changes"""
      if action==ObservableList.INSERT_ACTION:
        self.insert(items, position)
      elif action==ObservableList.REMOVE_ACTION:
        self.remove(items, position)
      elif action==ObservableList.MODIFY_ACTION:
        # some model items have been modified,
        # widget items are replaced by new items with new model items
        # from position, all items must be replaced with new value
        #remove and then insert new
        for modelItem in items:
          self.takeChild(position) # when an item is removed, position becomes the index of the next item
        self.insert(items, position)
      #else: print "unknown action"
      #print str(self.listView())

    def insert(self, items, position):
      """Insertion of items at position in the model
      -> create view items for all items and insert them at position in self
      insert at position = insert before item at position = insert after item at position-1"""
      i=position
      for item in items:
        if item.isLeaf(): # append item to keep the same order as in the model
          newItem=EditableTreeWidget.Leaf(None, item, None, self.iconDim)
        else: newItem=EditableTreeWidget.Branch(None, item, None, self.iconDim)
        self.insertChild(i, newItem)
        i+=1
        if item.unamed and self.treeWidget().hasFocus(): 
          self.editItem(newItem, 0)
          item.unamed=False
      self.setExpanded(True)
      #print "current item : ", self.treeWidget().currentItem()
      #print "set current item : ", newItem, newItem.text(0)
      self.treeWidget().setCurrentItem(newItem, 0)
      newItem.setSelected(True)
      
    def remove(self, items, position):
      """Remove items in the list from position
      if the list is empty, remove one item
      if position is undefined, remove items whose model is in the list"""
      if position is None:
        position=0
      if len(items)==0:
        self.takeChild(position)
      else:
        for modelItem in items:
          found=False
          i=position
          while not found and i<self.childCount():
            item=self.child(i)
            if item.model is modelItem:
              self.takeChild(i)
              found=True
            else: # if we remove an item, the index stay the same, no need to increment it
              i+=1

  #----------------------------------------------------------------------------
  class Leaf( Item ):
    """Item that represents a tree leaf. It doesn't have children."""
    def __init__( self, parent, treeItemModel, after=None, iconSize=defaultIconSize):
      EditableTreeWidget.Item.__init__(self, parent, treeItemModel, after, iconSize)
      #self.setExpandable(False)
      #self.setDropEnabled(False)
      self.setFlags( self.flags() & ~Qt.ItemIsDropEnabled )
    # This method could be used to change text color of items : 
    #def paintCell(self, painter, colorGroup, column, width, alignment ):
      #cg=qt.QColorGroup(colorGroup)
      #c=qt.QColor(cg.text())
      #cg.setColor( qt.QColorGroup.Text, Qt.red)
      #QListViewItem.paintCell(self,  painter, cg, column, width, alignment )
      #cg.setColor( qt.QColorGroup.Text, c )

#----------------------------------------------------------------------------
class EditableTreeController:
  """The controller is called to make changes on the model when events occur."""
  def __init__(self, m):
          self.model=m

  def drop(self, sourceTreeWidget, targetTreeWidget, draggedItems, draggedItemsParent, dropOn, dropAfter, dropBefore, copyMode):
    """
    Called when an item is dropped. It is copied or moved in target model.
    The copy of item will be modifiable even if source item is not.

    @type sourceTreeWidget: EditableTreeWidget
    @param sourceTreeWidget: the widget which dragged item comes from
    @type targetTreeWidget: EditableTreeWidget
    @param targetTreeWidget: the widget in which dragged item is dropped
    @type draggedItems: list of EditableTree.Item
    @param draggedItems: items to move or copy
    @type draggedItemsParent: list of EditableTree.Branch or EditableTree
    @param draggedItemsParent: container of dragged items
    @type dropOn: EditableTree.Branch or EditableTree
    @param dropOn: container in which item is dropped
    @type dropAfter: EditableTree.Item
    @param dropAfter: item after which dragged item is dropped
    @type dropBefore: EditableTree.Item
    @param dropBefore: item before which dragged item is dropped
    """
    insertedItem=None
    insertIndex=None
    if sourceTreeWidget==targetTreeWidget: # inner drop
      i=0
      for draggedItem in draggedItems:
        if copyMode: # if ctrl key is pressed, copy item, else move
          insertedItem, insertIndex=self.copy(draggedItem, dropOn, dropAfter, dropBefore, insertIndex)
        else:
          insertedItem, insertIndex=self.move(draggedItem, draggedItemsParent[i], dropOn, dropAfter, dropBefore, insertIndex)
        i+=1
        if insertedItem!=None:
          dropAfter=insertedItem
          dropBefore=None
          if insertIndex is not None:
            insertIndex+=1
    else: # drop from one widget to another 
      #print "drop on ", targetTreeWidget
      if not copyMode: sourceTreeWidget.deleteDraggedItems()
      for draggedItem in draggedItems:
        draggedItem.setAllModificationsEnabled(True)
        insertedItem, insertIndex=self.insert(draggedItem, dropOn, dropAfter, dropBefore, insertIndex)
        if insertedItem is not None:
          dropAfter=insertedItem
          dropBefore=None
          if insertIndex is not None:
            insertIndex+=1

  def move(self, item, source, target, after, before, index):
    """Move an element of the tree from source to target."""
    insertedItem=None
    insertIndex=index
    if not target.isDescendant(item): # can't move an item in its descendant
      # the item is deleted from its parent
      if source == target: # insertion index can change if the item is moved inside its container
        index=None
      self.delete(item, source)
      if item.copyEnabled and target.modifiable and not target.has_key(item.id): # if the item was not deletable, it is already in the target and can't be moved. 
        itemCopy = copy.deepcopy(item)
        # add this new item in the target item
        insertedItem, insertIndex=self.insert(itemCopy, target, after, before, index)
    return (insertedItem, insertIndex)

  def copy(self, item, target, after, before, index):
    """Recursive copy of an item in another item.
    A copy is completly modifiable, even if the source is not."""
    # create a deepcopy of the dragged item
    # notifiers are not copied, view items will be removed so they shouldn't be notified of changes in the new item
    insertedItem=None
    insertIndex=index
    if item.copyEnabled and target.modifiable:
      itemCopy = copy.deepcopy(item)
      # a copy is completely modifiable, even if the source is not
      itemCopy.setAllModificationsEnabled(True)
      # add this new item in the target item
      insertedItem, insertIndex=self.insert(itemCopy, target, after, before, index)
      #print self.model
    return (insertedItem, insertIndex)

  def insert(self, item, target, after, before, index):
    """Insert item in target, eventually after or before an existing child."""
    insertedItem=item
    insertIndex=index
    if not target.has_key(item.id): # can't copy an item that is already in the target
      if index is not None:
        target.insert(index, item.id, item)
      else:
        if before!=None:
          insertIndex=target.index(before.id)
          target.insert(insertIndex, item.id, item) # inserer le suivant en i+1
        elif after!=None: # insert the item after item "after"
          insertIndex=target.index(after.id)+1
          target.insert(insertIndex, item.id, item) # inserer le suivant en i+2
        else: # if no position, append the item
          #print "model append", self.model.onChangeNotifier._listeners
          target[item.id]=item # rien a preciser on ajoute a la fin
    else:
      newItem=target[item.id]
      if newItem is not after and newItem is not before: # else no move to do, it is already in the right place.
        insertedItem, insertIndex=self.move(newItem, target, target, after, before, None) 
      else: 
        insertedItem=newItem
        insertIndex=target.index(newItem.id)
    return (insertedItem, insertIndex)

  def delete(self, item, source):
    """Delete an item in the tree
    Delete item from source"""
    if item.delEnabled:
      del source[item.id]
      #print self.model

  def newItem(self, target):
    """Creates a new child in target"""
    if target.modifiable: #it is possible to add new item in the target
      # create a new branch item of the model
      # EditableTree class is not used directly because the model could be a derived class
      # class Branch should be able to be called without parameters
      newItem=self.model.__class__.Branch()
      target[newItem.id]=newItem

#----------------------------------------------------------------------------
class ObservableListWidget(QTreeWidget):
  """A widget to represent an ObservableList.
  The list is modifiable by manipulating items in the widget :
    - add new item
    - delete an item
    - rename an item

  The widget is created with an ObservableList as model.

  The graphic component has a reference on the model.
  It registers a callback method on the model notifier.
  On modification events, the model is updated and notify all its listeners,
  so the view is updated via updateContent method.

  Action events are not catched in this class.
  To enable actions, you may attach a popup menu to the signal contextMenuRequested.

  The widget shows tooltips items.
  All items icon are resized to have the same size.

  Inner classes :
    - Item

  Inherits qt component QTreeWidget.

  @type model: ObservableList with an attribute name
  @ivar model: the list which this widget is a graphic representation.
  @type tooltipsViewer: ListViewToolTip
  @ivar tooltipsViewer: this object reimplements QToolTip in order to show tooltip's item when mouse is on an item
  """

  def __init__(self, model, parent=None, iconSize=defaultIconSize):
    """
    @type model: ObservableList
    @param model: the list which this widget is a graphic representation.
    @type parent:QWidget
    @param parent: parent of this widget
    @type iconSize: couple of integers or None
    @param iconSize: force items icon resizing.
    """
    QTreeWidget.__init__( self, parent)
    self.setColumnCount(1)
    self.setHeaderLabels([""])
    self.iconDim=iconSize
    self.setIconSize(QSize(*iconSize))
    self.connect(self, SIGNAL("currentItemChanged (  QTreeWidgetItem *, QTreeWidgetItem * )", self.currentItemRenamed))
    self.setContextMenuPolicy(Qt.CustomContextMenu)
    # keep a reference to the model for control event and register a listener to be aware of changes
    self.setModel(model)

  def setModel(self, m):
    """Construct items of the list reading the model given.
    A listener is added to the model, so when the model change, the method self.updateContents
    is automatically called.
    """
    self.clear()
    self.model=m
    if m is not None:
      self.setHeaderLabels( [m.name] )
      self.model.addListener(self.updateContent)
      # create child items with data in the tree model
      lastChild=None
      for item in self.model:
        lastChild=ObservableListWidget.Item(self, item, lastChild, self.iconDim) # append item to keep the same order as in the model

  def currentItemRenamed(self, current, previous):
    if self.currentItem() is not None:
      self.currentItem().okRename()
    
  def getChild(self, n):
    """Return the nth child item
    If n<=0 return the first child, if n>=nb children, return the last child
    @rtype: ObservableListWidget.Item"""
    if n<0:
      n=0
    elif n>=self.topLevelItemCount():
      n=self.topLevelItemCount()-1
    return self.topLevelItem(n)

  def getLastChild(self):
    """Return the last child item.
    @rtype: ObservableListWidget.Item
    """
    return self.topLevelItem( self.topLevelItemCount() - 1)

  #------ Update on model changes ------
  def updateContent(self, action, items, position=None):
    """This method is called when the model notifies a change :
    The action has been done at position, with items.
    The view should update its content to reflect the changes.
    """
    if action==ObservableList.INSERT_ACTION:
      self.insert(items, position)
    elif action==ObservableList.REMOVE_ACTION:
      self.remove(items, position)
    elif action==ObservableList.MODIFY_ACTION:
      # some model items have been modified,
      # widget items are replaced by new items with new model items
      # from position, all items must be replaced with new value
      # ->remove and then insert new
      if position is None:
        position=0
      for modelItem in items:
        self.takeTopLevelItem(position)
      self.insert(items, position)
    #else: print action, "unknown action"

  def insert(self, items, position=None):
    """insertion of items at position in the model
      -> create view items for all items and insert them at position in self.
      Inserted item becomes the selected item."""
    #insert at position = insert before item at position = insert after item at position-1
    i=position
    for item in items:
      widgetItem=ObservableListWidget.Item(None, item, None, self.iconDim)
      self.insertTopLevelItem( i, item )
      i+=1
      if item.unamed:
        self.editItem(widgetItem)
        item.unamed=False
    self.setCurrentItem(widgetItem)

  def remove(self, items, position):
    """Removes items in the list from position.
    If the list is empty, one item is removed.
    If position is undefined, removes items whose model is in the list"""
    if position is None:
      position =0
    if len(items)==0:
      self.takeTopLevelItem(position)
    else:
      for modelItem in items:
        found=False
        i=position
        while not found and i<=self.topLevelItemcount():
          item=self.topLevelItem(i)
          if item.model is modelItem:
            self.takeTopLevelItem(i)
            found=True
          else:
            i+=1

  #----------------------------------------------------------------------------
  class Item(QTreeWidgetItem):
    """Item is the base class for elements of ObservableListWidget.
    Treats renaming events if the item is modifiable.
    """
    def __init__( self, parent, model, after=None, iconSize=defaultIconSize):
      """
      @type parent: ObservableListWidget
      @param parent: container of the item
      @type model: any object that contains attributes name, icon, tooltip and modifiable, and which is Observable.
      @param model: model which this item is the representation
      @type after: item
      @param after: the item after which current item must be added in the parent
      """
      QTreeWidgetItem.__init__( self, parent )
      if model.icon:
        iconPath = findIconFile( model.icon ) # QIcon
        if iconPath is not None:
          self.setIcon(0, QIcon(iconPath))
      self.setText(0, model.name)
      self.setToolTip(0, model.tooltip)
      # rename is enabled only if the item is modifiable
      if model.modifiable:
        self.setFlags(self.flags() | Qt.ItemIsEditable)
      self.model=model
      self.model.onAttributeChange("name", self.updateName)

    def getText(self):
      return self.text(0)

    def okRename(self):
      """This method is called when user renames the item.
      The name must be changed in the model."""
      if getattr(self, "model", None) is not None:
        newText=unicode(self.text(0))
        if self.model.name != newText:
          if self.model.name==self.model.tooltip:
            self.model.tooltip=newText
          self.model.name=newText

    #------ Update on model changes ------
    def updateName(self, newName):
      """This method is called when the model notifies that its name attribute has changed :
      The view should update its content to reflect the changes"""
      self.setText(0, newName)
      self.setToolTip(0, self.model.tooltip)
      #else: print "unknown action"
      
#----------------------------------------------------------------------------
class TreeListWidget(QTreeWidget):
  """
  This widget represents a list of EditableTree.
  The given model is an ObservableSortedDictionary (trees are referenced by their id).
  
  The widget listens for modifications done on the model :
    - add new item
    - delete an item
    - rename an item
  It registers a callback method on the model notifier.
  On modification events, the model is updated and notify all its listeners,
  so the view is updated via updateContent method.

  Action events are not catched in this class.
  To enable actions, you may attach a popup menu to the signal contextMenuRequested.
  Items accept drop if dropped elements are EditableTree.Item. Dropped element are added in target EditableTree.
  
  The widget shows tooltips items.
  All items icon are resized to have the same size.
  
  Inner classes :
    - Item

  Inherits qt component QTreeWidget.

  @type model: ObservableSortedDictionary
  @ivar model: the EditableTree map which this widget is a graphic representation.
  @type tooltipsViewer: ListViewToolTip
  @ivar tooltipsViewer: this object reimplements QToolTip in order to show tooltip's item when mouse is on an item
  @type draggedItems: list of EditableTree.Item
  @ivar draggedItems: list of elements that are currently dragged on the widget
  @type dropOn: TreeListWidget.Item
  @ivar dropOn: current drop zone (mouse is over this item with dragged items and the drop zone item accepts drop)
  """

  def __init__(self, model, parent=None, iconSize=defaultIconSize):
    """
    @type model: ObservableSortedDictionary
    @ivar model: the EditableTree map which this widget is a graphic representation.
    @type parent:QWidget
    @param parent: parent of this widget
    @type iconSize: couple of integers or None
    @param iconSize: force items icon resizing.
    """
    QTreeWidget.__init__( self, parent)
    self.setColumnCount(1)
    self.iconDim=iconSize
    self.setIconSize(QSize(*iconSize))
    self.setRootIsDecorated(False) # it is not really a tree but a list, keep a QTreeWidget to keep columns
    # enable display of tooltips on items
    # the listview accept drops to enable adding items in trees by dropping items on the item representing the tree
    self.setAcceptDrops(True)
    self.draggedItems=[]
    self.dropOn=None
    self.connect(self, SIGNAL("currentItemChanged ( QTreeWidgetItem *, QTreeWidgetItem * )"), self.currentItemRenamed )
    self.setContextMenuPolicy(Qt.CustomContextMenu)
    # keep a reference to the model for control event and register a listener to be aware of changes
    self.setModel(model)

  def setModel(self, m):
    """
    Construct items of the list reading the model given.
    A listener is added to the model, so when the model change, the method self.updateContents
    is automatically called.
    @type m: ObservableSortedDictionary
    @param m: tree map to show in this widget
    """
    self.clear()
    self.model=m
    if m is not None:
      self.setHeaderLabels([m.name])
      self.model.addListener(self.updateContent)
      # create child items with data in the tree model
      lastChild=None
      for item in self.model.values():
        lastChild=self.Item(self, item, lastChild, self.iconDim) # append item to keep the same order as in the model

  def currentItemRenamed(self, current, previous):
    if self.currentItem() is not None:
      self.currentItem().okRename()
    
  def getChild(self, n):
    """
    Return the nth child item
    If n<=0 return the first child, if n>=nb children, return the last child
    @rtype: TreeListWidget.Item
    """
    if n<0:
      n=0
    elif n>=self.topLevelItemCount():
      n=self.topLevelItemCount() - 1
    return self.topLevelItem(n)

  def getLastChild(self):
    """
    Return the last child item.
    @rtype: TreeListWidget.Item
    """
    return self.topLevelItem(self.topLevelItemCount()-1)
  
  #------ Drag&drop control  ------
  def dragEnterEvent(self, event):
    """
    This method is called when a drag enters in the widget. 
    The event contains the encoding of the dragged objects.
    The method tries to decode text event as minf representation of EditableTree.Item and sets self.draggedItems.
    """
    # accept drag if it contains instances of EditableTree.Item
    self.draggedItems=[]
    if event.mimeData().hasText():
      textEventBuf=StringIO(event.mimeData().text())
      # check if it is the expected minf format
      format, reduction=minfFormat(textEventBuf)
      if format=="XML":
        textEventBuf.seek(0)
        for value in iterateMinf( textEventBuf ):
          if isinstance(value, EditableTree.Item):
            event.accept()
            value.setAllModificationsEnabled(True) # this is a copy of items that comes from another tree, so enable modifications on the copied items
            self.draggedItems.append(value)
          else: event.ignore()
        textEventBuf.close()
      else: event.ignore()
    else: event.ignore()
  
  def toContentsPoint(self, point):
    """
    Translates coordinates in the frame to coordinates in the content of the list (without header).
    @rtype: QPoint
    """
    #viewportPoint=self.contentsToViewport(point) to get the item at position we need the position in the entire list not in the visible content (viewport)
    return point #- QPoint(0, self.header().height())
 
  def dragMoveEvent(self, event):
    """
    When the drag moves, the current dropzone can change.
    According to cursor position, dropOn attribute is updated.
    To have a valid dropzone, mouse must be over a modifiable item.
    When the dragMoveEvent is ignored, a forbidden cursor is shown.
    """
    # draggedItem must be set, otherwise it isn't a correct drag 
    if self.draggedItems==[]:
      event.ignore(self.contentsRect())
      self.dropOn=None
    else:
      cursorPos=self.toContentsPoint(event.pos())
      currentItem=self.itemAt(cursorPos) #item under the mouse cursor
      self.dropOn=currentItem
      if currentItem is not None and currentItem.model.modifiable and event.source().model is not currentItem.model: # not accept dragged items that comes from current tree
        event.accept()
      else: event.ignore()
 
  def dropEvent( self, event):
    """
    When items are dropped, they are added to the dropzone model (an EditableTree)
    """
    # current dropzone is stored in dropOn
    if self.dropOn!=None:
      for item in self.draggedItems:
        self.dropOn.model.add(item)
    self.draggedItems = []
    self.dropOn=None
    
  #------ Update on model changes ------
  def updateContent(self, action, items, position=None):
    """
    This method is called when the model notifies a change :
    The action has been done at position, with items.
    The view should update its content to reflect the changes.
    """
    if action==ObservableList.INSERT_ACTION:
      self.insert(items, position)
    elif action==ObservableList.REMOVE_ACTION:
      self.remove(items, position)
    elif action==ObservableList.MODIFY_ACTION:
      # some model items have been modified,
      # widget items are replaced by new items with new model items
      # from position, all items must be replaced with new value
      # ->remove and then insert new
      for modelItem in items:
        self.takeTopLevelItemItem(position)
      self.insert(items, position)
    #else: print action, "unknown action"

  def insert(self, items, position=None):
    """insertion of items at position in the model
      -> create view items for all items and insert them at position in self.
      Inserted item becomes the selected item."""
    #insert at position = insert before item at position = insert after item at position-1
    i=position
    for item in items:
      newItem=self.Item(None, item, None, self.iconDim)
      self.insertTopLevelItem(i, newItem)
      i+=1
      if item.unamed:
        self.editItem(newItem, 0)
        item.unamed=False
    self.setCurrentItem(newItem)

  def remove(self, items, position):
    """Removes items in the list from position.
    If the list is empty, one item is removed.
    If position is undefined, removes items whose model is in the list"""
    if position is None:
      position =0
    if len(items)==0:
      self.takeTopLevelItem(position)
    else:
      for modelItem in items:
        found=False
        i=position
        while not found and i<self.count():
          item=self.topLevelItem(i)
          if item.model is modelItem:
            self.takeTopLevelItem(i)
            found=True
          else:
            i+=1

  #----------------------------------------------------------------------------
  class Item(QTreeWidgetItem):
    """
    Item is the base class for elements of TreeListWidget.
    Treats renaming events if the item is modifiable.
    """
    def __init__( self, parent, model, after=None, iconSize=defaultIconSize):
      """
      @type parent: TreeListWidget
      @param parent: container of the item
      @type model: any object that contains attributes name, icon, tooltip and modifiable, and which is Observable.
      @param model: model which this item is the representation
      @type after: item
      @param after: the item after which current item must be added in the parent
      """
      QTreeWidgetItem.__init__( self, parent )
      if model.icon:
        iconPath = findIconFile( model.icon )
        if iconPath is not None:
          self.setIcon(0, QIcon(iconPath))
      self.setText(0, model.name)
      self.setToolTip(0, model.tooltip)
      # rename is enabled only if the item is modifiable
      if model.modifiable:
        self.setFlags(self.flags() | Qt.ItemIsEditable)
      if not model.visible:
        self.setHidden(True)
      self.model=model
      self.model.onAttributeChange("name", self.updateName)
      self.model.onAttributeChange("visible", self.updateVisibility)

    def getText(self):
      return self.text(0)

    def okRename(self):
      """
      This method is called when user renames the item.
      The name must be changed in the model.
      """
      if getattr(self, "model", None) is not None:
        newText=unicode(self.text(0))
        if self.model.name != newText:
          if self.model.name==self.model.tooltip:
            self.model.tooltip=newText
          self.model.name=newText # the model will notify the change and updateName will be called

    #------ Update on model changes ------
    def updateName(self, newName):
      """
      This method is called when the model notifies that its name attribute has changed :
      The view should update its content to reflect the changes.
      """
      self.setText(0, newName)
      self.setToolTip(0, self.model.tooltip)
      #else: print "unknown action"
      
    def updateVisibility(self, newValue):
      """
      Called when the model notifies that its visible attribute value has changed.
      """
      #pass
      self.setHidden(not newValue)



# -*- coding: utf-8 -*-
import json
import os
import collections 
import datetime
import glob
from soma.sorted_dictionary import SortedDictionary
from soma.controller import Controller
try:
    from traits.api import HasTraits,Str,Enum,Directory,File
except ImportError:
    from enthought.traits.api import HasTraits,Str,Enum,Directory,File
from soma.application import Application

class Study(Controller):
    _instance=None
    """Class to write and save informations about process in the json"""
    def __init__(self):
        super(Study, self).__init__()
        HasTraits.__init__(self) 
        # Find foms available
        foms = Application().fom_manager.find_foms()
        foms.insert(0,' ')
        self.add_trait('input_directory',Directory)
        #self.add_trait('input_directory',Directory('/nfs/neurospin/cati/cati_shared'))
        self.add_trait('input_fom',Enum(foms))
        self.add_trait('output_directory',Directory)   
        self.add_trait('output_fom',Enum(foms))
        self.add_trait('shared_directory',Directory)
        self.add_trait('spm_directory',Directory)
        self.add_trait('format_image',Str)
        self.add_trait('format_mesh',Str)
        self.add_trait('process',Enum(
            '   ',
            'morphologist.process.morphologist_simplified.SimplifiedMorphologist',
            'morphologist.process.morphologist.Morphologist',
            'morpho.morphologist.morphologist',
            'morphologist.process.customized.morphologist.CustomMorphologist'))
        #self.process_specific=None
        self.compteur_run_process={}
        self.runs=collections.OrderedDict()
        
    #def _input_fom_changed(self, old, new):
        #print 'input fom changed'
        ##Get list format
        #self.add_trait(œ'coucou',Str)
        
        
    @staticmethod
    def get_instance():
        if Study._instance is None:
            Study._instance=Study()
            return Study._instance
        else:
            return Study._instance
        
    """Save on json with OrderedDict"""        
    def save(self):
        self.name_study=str(self.output_directory.split(os.sep)[-1])
        self.dico=collections.OrderedDict([('name_study',self.name_study),('input_directory',self.input_directory),('input_fom',self.input_fom),('output_directory',self.output_directory),('output_fom',self.output_fom),('shared_directory',self.shared_directory),('spm_directory',self.spm_directory),('format_image',self.format_image),('format_mesh',self.format_mesh),('process',self.process)])
        json_string = json.dumps(self.dico, indent=4, separators=(',', ': '))
        with open(os.path.join(self.output_directory,self.name_study+'.json'), 'w') as f:
            f.write(unicode(json_string))

    """Load and put on self.__dict__ OrderedDict"""        
    def load(self,name_json):
        try:
            with open(name_json, 'r') as json_data:
                self.__dict__ = json.load(json_data,object_pairs_hook=collections.OrderedDict)        
            for element in self.__dict__:
                setattr(self,element,self.__dict__[element])
        
        #No file to load
        except IOError:
            pass
        
        
    #"""Get number of run process and iterate"""        
    #def inc_nb_run_process(self,process_name):
        #print 'process_name',process_name
        #if self.compteur_run_process.has_key(process_name):
            #valeur=self.compteur_run_process[process_name]
            #print 'valeur',valeur
            #self.compteur_run_process[process_name]=valeur+1
        #else:
            #self.compteur_run_process[process_name]=1


    def save_run(self,attributes,process):
        print 'save run'
        #Create date directory
        date=datetime.datetime.now()
        date_directory=str(date.day)+'_'+str(date.month)+'_'+str(date.year)
        directory=os.path.join(self.output_directory,date_directory)
        if not os.path.exists(directory):
            name_run=process.name+str(1)
            os.makedirs(directory)
            self.compteur_run_process[process.name]=1
        
        else:
            list_json=glob.glob(directory+os.sep+'*.json')
            inc=1
            number=len(list_json)+inc
            name_run=process.name+str(number)
            while os.path.exists(os.path.join(directory,name_run+'.json')) is True:
                inc=inc+1
                number=len(list_json)+inc
                name_run=process.name+str(number)
            self.compteur_run_process[process.name]=number
        
        run=collections.OrderedDict()
        run['process_name']='morphologist.process.morphologist_simplified.SimplifiedMorphologist'
        #self.inc_nb_run_process(process.name )
        run['attributes']={}
        for key in attributes:
            run['attributes'][key]=attributes[key]
        run['parameters']=collections.OrderedDict()
        run['input']=collections.OrderedDict()
        run['output']=collections.OrderedDict()
        #dicti_sorted=SortedDictionary(*[(key,a[key]) for key in a])

        for name, trait in process.user_traits().iteritems():
            if trait.is_trait_type( File ) is False:
                run['parameters'][name]=getattr(process,name)
            elif trait.output is False:
                run['input'][name]=getattr(process,name)
            elif trait.output is True:
                run['output'][name]=getattr(process,name)

        dico=collections.OrderedDict([('name_study',self.name_study),('input_directory',self.input_directory),('input_fom',self.input_fom),('output_directory',self.output_directory),('output_fom',self.output_fom),('shared_directory',self.shared_directory),('spm_directory',self.spm_directory),('format_image',self.format_image),('format_mesh',self.format_mesh),('run',run)])
        json_string = json.dumps(dico, indent=4, separators=(',', ': '))
        
        with open(os.path.join(directory,name_run+'.json'), 'w') as f:
            f.write(unicode(json_string)) 

        #To incremente compteur runs
        self.save()



# -*- coding: iso-8859-1 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Reading of XML minf format.

@author: Yann Cointepas
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''
__docformat__ = "epytext en"

import sys, os, operator, codecs, types, gzip
from xml.sax.saxutils import quoteattr as xml_quoteattr
from xml.sax.saxutils import escape as xml_escape
from xml.sax import make_parser
from xml.sax.handler import ContentHandler
from xml.sax.handler import ErrorHandler

from soma.translation import translate as _
from soma.undefined import Undefined
from soma.minf.error import MinfError

from soma.minf.reader import MinfReader
from soma.minf.tree import minfStructure, listStructure, dictStructure, \
                            StartStructure, EndStructure, Reference
from soma.minf.xhtml import XHTML

# This module only contains a definition of XML tags and attributes.
# It is designed to allow "import *".
from soma.minf.xml_tags import *
from cStringIO import StringIO

#------------------------------------------------------------------------------
class MinfXMLReader( MinfReader, ErrorHandler ):
  '''
  Specialization of L{MinfReader} class for reading XML minf format.
  '''
  name = 'XML'
  
  class SaxHandler( ContentHandler ):
    def __init__( self, parser ):
      self.parser = parser
      self.parser._handler = MinfXMLHandler( parser )
      self.parser._stack = []
      
    def startElement( self, name, attrs ):
      # Convert to builtin Python types
      name = unicode( name )
      attrs = dict( [ (unicode(i[0]),unicode(i[1])) for i in attrs.items()  ] )
      self.parser._stack.append( name )
      self.parser._handler.startElement( self.parser, name, attrs )
    
    
    def endElement( self, name ):
      self.parser._handler.endElement( self.parser, unicode( name ) )
      self.parser._stack.pop()
    
    
    def characters( self, content ):
      self.parser._handler.characters( self.parser, unicode( content ) )
  
  
  def reduction( self, source ):
    self._nodesToProduce = []
    self._sax_parser = make_parser()
    self._sax_parser.setContentHandler( MinfXMLReader.SaxHandler( self ) )
    self._sax_parser.setErrorHandler( self )
    self._sax_parser.reset()
    self._minfStarted = False
    self._minfFinished = False
    self._nodeIdentifier = 0
    buffer = []
    while not self._nodesToProduce:
      line = source.readline()
      buffer.append( line )
      if not line:
        # end of file
        break
      self._sax_parser.feed( line )
    if self._nodesToProduce:
      return ( self._nodesToProduce[ 0 ].attributes[ 'reduction' ], ''.join(buffer) )
    return ( None, ''.join(buffer) )


  def nodeIterator( self, source ):
    self._nodesToProduce = []
    self._sax_parser = make_parser()
    self._sax_parser.setContentHandler( MinfXMLReader.SaxHandler( self ) )
    self._sax_parser.setErrorHandler( self )
    self._sax_parser.reset()
    #self._sax_parser.feed( encodingLine )
    self._minfStarted = False
    self._minfFinished = False
    self._nodeIdentifier = 0
    while not self._minfFinished:
      while self._nodesToProduce:
        node = self._nodesToProduce.pop( 0 )
        yield node
      line = source.readline()
      if not line:
        # end of file
        if not self._minfStarted:
          # Log files may be read while not finished, therefore the closing
          # tag is missing. It is append to avoid XML parser error message.
          self._sax_parser.feed( '<' + minfTag + '/>' )
        self._minfFinished = True
        break
      self._sax_parser.feed( line )
    while self._nodesToProduce:
      node = self._nodesToProduce.pop( 0 )
      yield node


  def parseError( self, errorMessage ):
    self.fatalError( errorMessage )
  
  def fatalError( self, error ):
    raise MinfError( _( 'XML parse error: %s' ) % \
                        ( unicode( error ), )  + ' (stack = ' + \
        ','.join( [ '"' + i + '"' for i in self._stack ]) + ')' )


  def checkNoMoreAttributes( self, attributes ):
    if attributes:
      self.parseError( _('Invalid attribute": %(attributes)s') %\
        { 'attributes': ', '.join( ['"' + n + '"' for n in attributes]  ) } )
  

#------------------------------------------------------------------------------
class XMLHandler:
  def __init__( self, parser, parent, name, attributes ):
    self.parent = parent
    parser.checkNoMoreAttributes( attributes )
    
  def startElement( self, parser, name, attributes ):
    parser.parseError( _( 'Unexpected tag "%s"') % ( name, ) )


  def endElement( self, parser, name ):
    parser._handler = self.parent
    self.parent = None
  
  
  def characters( self, parser, content ):
    pass


#------------------------------------------------------------------------------
class NoneXMLHandler( XMLHandler ):
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( None )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class TrueXMLHandler( XMLHandler ):
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( True )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class FalseXMLHandler( XMLHandler ):
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( False )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class NumberXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    XMLHandler.__init__( self, parser, parent, name, attributes )
    self._stringValue = []
  
  
  def characters( self, parser, content ):
    self._stringValue.append(content)
  
  
  def endElement( self, parser, name ):
    stringValue=''.join(self._stringValue)
    try: 
      value = int( stringValue )
    except:
      try: 
        value = long( stringValue )
      except:
        value = float( stringValue )
    parser._nodesToProduce.append( value )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class StringXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    XMLHandler.__init__( self, parser, parent, name, attributes )
    self._value = []
  
  
  def characters( self, parser, content ):
    self._value.append(content)
  
  
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( ''.join(self._value) )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class ListXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    length = attributes.pop( lengthAttribute, None )
    identifier = attributes.pop( identifierAttribute, None )
    if length is not None:
      parser._nodesToProduce.append( StartStructure( listStructure, 
                                                     identifier=identifier,
                                                     length=length ) )
    else:
      parser._nodesToProduce.append( StartStructure( listStructure,
                                                     identifier=identifier ) )
    XMLHandler.__init__( self, parser, parent, name, attributes )

  
  def startElement( self, parser, name, attributes ):
    newHandler = MinfXMLHandler.getHandler( parser, self, name, attributes )
    if newHandler is None:
      parser.parseError( _( 'Unexpected tag "%s"' ) % ( name, ) )
    parser._handler = newHandler
  
  
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( EndStructure( listStructure ) )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class DictionaryXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    length = attributes.pop( lengthAttribute, None )
    identifier = attributes.pop( identifierAttribute, None )
    if length is not None:
      parser._nodesToProduce.append( StartStructure( dictStructure,
                                                     identifier=identifier, 
                                                     length=length ) )
    else:
      parser._nodesToProduce.append( StartStructure( dictStructure,
                                                     identifier=identifier ) )
    XMLHandler.__init__( self, parser, parent, name, attributes )

  
  def startElement( self, parser, name, attributes ):
    nameAttr = attributes.pop( nameAttribute, None )
    if nameAttr is not None:
      parser._nodesToProduce.append( nameAttr )
    newHandler = MinfXMLHandler.getHandler( parser, self, name, attributes )
    if newHandler is None:
      parser.parseError( _( 'Unexpected tag "%s"' ) % ( name, ) )
    parser._handler = newHandler
  
  
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( EndStructure( dictStructure ) )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class FactoryXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    self.type = attributes.pop( objectTypeAttribute, None )
    if self.type is None:
      parser.parseError( _( '%s attribute missing' ) % ( objectTypeAttribute, ) )
    identifier = attributes.pop( identifierAttribute, None )
    parser._nodesToProduce.append( StartStructure( self.type, identifier=identifier ) )
    XMLHandler.__init__( self, parser, parent, name, attributes )
  
  
  def startElement( self, parser, name, attributes ):
    nameAttr = attributes.pop( nameAttribute, None )
    parser._nodesToProduce.append( nameAttr )
    newHandler = MinfXMLHandler.getHandler( parser, self, name, attributes )
    if newHandler is None:
      parser.parseError( _( 'Unexpected tag "%s"' ) % ( name, ) )
    parser._handler = newHandler
  
  
  def endElement( self, parser, name ):
    parser._nodesToProduce.append( EndStructure( self.type ) )
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class ReferenceXMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    identifier = attributes.pop( identifierAttribute, None )
    if self.type is None:
      parser.parseError( _( '%s attribute missing' ) % ( identifierAttribute, ) )
    parser._nodesToProduce.append( Reference( identifier=identifier ) )
    XMLHandler.__init__( self, parser, parent, name, attributes )
  
  
  def endElement( self, parser, name ):
    XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class XHTMLHandler( XMLHandler ):
  def __init__( self, parser, parent, name, attributes ):
    self.stack = [ XHTML( name, attributes ) ]
    self._characters = []
    XMLHandler.__init__( self, parser, parent, name, attributes )


  def characters( self, parser, content ):
    self._characters.append(content)
  
  
  def startElement( self, parser, name, attributes ):
    c = ''.join(self._characters)
    if c: self.stack[ -1 ].content.append( c )
    self._characters = []
    newItem = XHTML( name, attributes )
    self.stack[ -1 ].content.append( newItem )
    self.stack.append( newItem )
    
    
  def endElement( self, parser, name ):
    item = self.stack.pop()
    c = ''.join(self._characters)
    self._characters = []
    if c: item.content.append( c )
    if not self.stack:
      parser._nodesToProduce.append( item )
      XMLHandler.endElement( self, parser, name )


#------------------------------------------------------------------------------
class MinfXMLHandler( XMLHandler ):
  _handlers = {
    noneTag: NoneXMLHandler,
    trueTag: TrueXMLHandler,
    falseTag: FalseXMLHandler,
    numberTag: NumberXMLHandler,
    stringTag: StringXMLHandler,
    listTag: ListXMLHandler,
    dictionaryTag: DictionaryXMLHandler,
    factoryTag: FactoryXMLHandler,
    referenceTag: ReferenceXMLHandler,
    xhtmlTag: XHTMLHandler,
  }
  
  def __init__( self, parser ):
    XMLHandler.__init__( self, parser, None, minfTag, {} )
  
  
  def getHandler( parser, parent, name, attributes ):
    handler = MinfXMLHandler._handlers.get( name )
    if handler is not None:
      handler = handler( parser, parent, name, attributes )
    return handler
  getHandler = staticmethod( getHandler )
  
  
  def startElement( self, parser, name, attributes ):
    if parser._minfStarted:
      nameAttr = attributes.pop( nameAttribute, None )
      if nameAttr is None:
        if self._obsoleteFormat:
          parser.parseError( _( '%s attribute required for minf_1.0' ) % ( nameAttribute,) )
      else:
        if self._obsoleteFormat:
          parser._nodesToProduce.append( nameAttr )
        else: 
          parser.parseError( _( 'Unexpected attribute %s' ) % ( nameAttribute, ) )
      newHandler = MinfXMLHandler.getHandler( parser, self, name, attributes )
      if newHandler is None:
        parser.parseError( _( 'Unexpected tag "%s"' ) % ( name, ) )
      parser._handler = newHandler
    else:
      if  name != minfTag:
        # Document is not a minf file
        parser.parseError(
          _( 'Wrong document type, expecting "%(minf)s" instead of ' \
                '"%(other)s>"' ) % { 'minf': minfTag, 'other': name } )
      # Checking minf format
      self._obsoleteFormat = False
      expanderName = attributes.pop( expanderAttribute, None )
      if expanderName is None:
        # Compatibility with obsolete minf 1.0 XML format
        version = attributes.pop( 'version', None )
        if version is None:
          expanderName = 'minf_2.0'
        else:
          if version != '1.0':
            parser.parseError( _( 'Wrong value for attribute "version", found '
                                    '"%s" but only "1.0" is accepted' ) % \
                              ( version, ) )
          self._obsoleteFormat = True
          expanderName = 'minf_1.0'
      parser._nodesToProduce.append( StartStructure( minfStructure, reduction=expanderName ) )
      # Compatibility with obsolete minf 1.0 XML format
      if self._obsoleteFormat:
        parser._nodesToProduce.append( StartStructure( dictStructure ) )
      parser._minfStarted = True
      # Check attributes
      parser.checkNoMoreAttributes( attributes )
  
  
  def endElement( self, parser, name ):
    if  name == minfTag:
      # Compatibility with obsolete minf 1.0 XML format
      if self._obsoleteFormat:
        parser._nodesToProduce.append( EndStructure( dictStructure ) )
      parser._nodesToProduce.append( EndStructure( minfStructure ) )
      parser._minfFinished = True
    else:
      parser._nodesToProduce.append( EndStructure( minfStructure ) )




# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility classes and functions for Python callable.

- author: Yann Cointepas
- organization: `NeuroSpin <http://www.neurospin.org>`_ and 
  `IFR 49 <http://www.ifr49.org>`_
- license: `CeCILL version 2 <http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>`_
'''
__docformat__ = 'restructuredtext en'

import inspect
from itertools import izip


#-------------------------------------------------------------------------------
from soma.translation import translate as _


#-------------------------------------------------------------------------------
class Empty( object ):
  pass

#-------------------------------------------------------------------------------
class SomaPartial( object ):
  '''
  Python 2.5 introduced a very useful function: :py:func:`functools.partial`,
  this is an implementation that is compatible with Python 2.3 (if
  functools.partial exists, it is used directly).
  
  functools.partial allow to create a new function from an existing
  function by setting values to some arguments. The new function
  will be callable with less parameters. See Python 2.5 documentation
  for more information.
  
  Example::

    from soma.functiontools import partial
    
    def f( a, b, c ):
      return ( a, b, c )
    
    g = partial( f, 'a', c='c' )
    g( 'b' ) # calls f( 'a', 'b', c='c' )
  '''
  def __init__( self, function, *args, **kwargs ):
    self.func = function
    self.args = args
    self.keywords = kwargs
  
  
  def __call__( self, *args, **kwargs ):
    merged_kwargs = self.keywords.copy()
    merged_kwargs.update( kwargs )
    return self.func( *(self.args + args), **merged_kwargs )
  
  @property
  def func_code( self ):
    '''
    This property make SomaPartial useable with traits module. The method
    on_trait_change is requiring that the function has a func_code.co_argcount
    attribute.
    '''
    result = Empty()
    result.co_argcount = numberOfParameterRange( self )[ 0 ]
    return result
    
try:
  from functools import partial
except ImportError:
  partial = SomaPartial

#-------------------------------------------------------------------------------
def getArgumentsSpecification( callable ):
  '''
  This is an extension of Python module :py:mod:`inspect.getargspec` that 
  accepts classes and returns only information about the parameters that can 
  be used in a call to *callable* (*e.g.* the first *self* parameter of bound 
  methods is ignored). If *callable* has not an appropriate type, a 
  :py:class:`TypeError` exception is raised.

  - callable: *function*, *method*, *class* or *instance*

    callable to inspect

  - returns: *tuple* of four elements
  
    As :py:func:`inspect.getargspec`, returns 
    *(args, varargs, varkw, defaults)* where *args* is a list of the argument
    names (it may contain nested lists). *varargs* and *varkw* are the names 
    of the ``*`` and ``**`` arguments or *None*. *defaults* is an n-tuple of 
    the default values of the last *n* arguments.
  '''
  if inspect.isfunction( callable ):
    return inspect.getargspec( callable )
  elif inspect.ismethod( callable ):
    args, varargs, varkw, defaults = inspect.getargspec( callable )
    args = args[ 1: ] # ignore the first "self" parameter
    return args, varargs, varkw, defaults
  elif inspect.isclass( callable ):
    try:
      init = callable.__init__
    except AttributeError:
      return [], None, None, None
    return getArgumentsSpecification( init )
  elif isinstance( callable, (partial,SomaPartial) ):
    args, varargs, varkw, defaults = getArgumentsSpecification( callable.func )
    if defaults:
      d = dict( izip( reversed( args ), reversed( defaults ) ) )
    else:
      d = {}
    d.update( izip( reversed( args ), reversed( callable.args ) ) )
    if callable.keywords:
      d.update( callable.keywords )

    if len(d) :
      defaults = tuple( (d[i] for i in args[ -len(d): ] ) )
    else :
      defaults = d
      
    return ( args, varargs, varkw, defaults )
  else:
    try:
      call = callable.__call__
    except AttributeError:
      raise TypeError( _( '%s is not callable' ) % \
                          repr( callable ) )
    return getArgumentsSpecification( call )

#-------------------------------------------------------------------------------
def getCallableString( callable ):
  '''
  Returns a translated human readable string representing a callable.
  
  - callable: *function*, *method*, *class* or *instance*
  
    callable to inspect
  
  - returns: *string*
  
    type and name of the callable
  '''
  if inspect.isfunction( callable ):
    name = _( 'function %s' ) % ( callable.__name__, )
  elif inspect.ismethod( callable ):
    name = _( 'method %s' ) % ( callable.im_class.__name__ + '.' + \
                                callable.__name__, )
  elif inspect.isclass( callable ):
    name = _( 'class %s' ) % ( callable.__name__, )
  else:
    name = str( callable )
  return name

#-------------------------------------------------------------------------------
def hasParameter( callable, parameterName ):
  '''
  Return True if *callable* can be called with a parameter named
  *parameterName*. Otherwise, returns *False*.
  
  - callable: *function*, *method*, *class* or *instance*
  
    callable to inspect
  
  - parameterName: *string*
  
    name of the parameter
  
  - returns: *bool*
  
  see: :py:func:`getArgumentsSpecification`
  '''
  args, varargs, varkw, defaults = getArgumentsSpecification( callable )
  return varkw is not None or parameterName in args

#-------------------------------------------------------------------------------
def numberOfParameterRange( callable ):
  '''
  Return the minimum and maximum number of parameter that can be used to call a
  function. If the maximum number of argument is not defined, it is set to 
  None.
  
  - callable: *function*, *method*, *class* or *instance*
  
    callable to inspect
  
  - returns: *tuple* of two elements
  
    (minimum, maximum)
  
  see: :py:func:`getArgumentsSpecification`
  '''
  args, varargs, varkw, defaults = getArgumentsSpecification( callable )
  if defaults is None or len( defaults ) > len( args ):
    lenDefault = 0
  else:
    lenDefault = len( defaults )
  minimum = len( args ) - lenDefault
  if varargs is None:
    maximum = len( args )
  else:
    maximum = None
  return minimum, maximum


#-------------------------------------------------------------------------------
def checkParameterCount( callable, paramCount ):
  '''
  Check that a callable can be called with *paramCount* arguments. If not, a
  RuntimeError is raised.
  
  - callable: *function*, *method*, *class* or *instance*
  
    callable to inspect
  
  - paramCount: *integer*
  
    number of parameters
  
  see: :py:func:`getArgumentsSpecification`
  '''
  minimum, maximum = numberOfParameterRange( callable )
  if ( maximum is not None and paramCount > maximum ) or \
     paramCount < minimum:
    raise RuntimeError( \
      _( '%(callable)s cannot be called with %(paramCount)d arguments' ) % \
      { 'callable': getCallableString( callable ), 
        'paramCount': paramCount }  )

#-------------------------------------------------------------------------------
def drange(start, stop, step = 1):
    '''
    Creates lists containing arithmetic progressions of any number type (integer, float, ...)
    '''
    r = start
    while r < stop:
        yield r
        r += step



# -*- coding: iso-8859-1 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility classes and functions for debugging.

- author: Yann Cointepas
- organization: `NeuroSpin <http://www.neurospin.org>`_ and 
  `IFR 49 <http://www.ifr49.org>`_
- license: `CeCILL version 2 <http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>`_
'''
from __future__ import absolute_import

__docformat__ = 'restructuredtext en'

import sys
from pprint import pprint

from soma.undefined import Undefined

def function_call_info( frame=None ):
  '''
  Return a dictionary that gives information about a frame corresponding to a function call.
  The directory contains the following items:

  - 'function': name of the function called
  - 'filename': name of the python file containing the function
  - 'lineno': line number executed in 'filename'
  - 'arguments': arguments passed to the function. It is a list containing 
    pairs of (argument name, argument value).
  '''
  try:
    if frame is None:
      frame = sys._getframe( 1 )
    result = { 
      'function': frame.f_code.co_name,
      'lineno': frame.f_lineno,
      'filename': frame.f_code.co_filename,
    }
    args = frame.f_code.co_varnames[:frame.f_code.co_argcount]
    result[ 'arguments' ] = \
      [( p, frame.f_locals.get( p, frame.f_globals.get( p, Undefined ) ) ) for p in args]
  finally:
    del frame
  return result


def stack_calls_info( frame=None ):
  '''
  Return a list containing function_call_info(frame) for all frame in the stack.
  '''
  try:
    if frame is None:
      frame = sys._getframe( 1 )
    result = []
    while frame is not None:
      result.insert( 0, function_call_info( frame ) )
      frame = frame.f_back
    return result
  finally:
    del frame


def print_stack( out=sys.stdout, frame=None ):
  '''
  Print information about the stack, including argument passed to functions called.
  '''
  try:
    if frame is None:
      frame = sys._getframe( 1 )
    for info in stack_calls_info( frame ):
      print >> out, 'File "%(filename)s", line %(lineno)d' % info + ' in ' + info[ 'function' ]
      for name, value in info['arguments']:
        out.write( '   ' + name + ' = ' )
        pprint( value, out, 3 )
    out.flush()
  finally:
    del frame



# -*- coding: iso-8859-1 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Writing of XML minf format.

@author: Yann Cointepas
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''
__docformat__ = "epytext en"

import codecs
from xml.sax.saxutils import quoteattr as xml_quoteattr
from xml.sax.saxutils import escape as xml_escape
from soma.translation import translate as _
from soma.minf.tree import createMinfReducer
from soma.minf.writer import MinfWriter
from soma.minf.tree import minfStructure, listStructure, dictStructure, \
                            StartStructure, EndStructure
from soma.minf.error import MinfError
from soma.undefined import Undefined


# This module only contains a definition of XML tags and attributes.
# It is designed to allow "import *".
from soma.minf.xml_tags import *

#: Replacement table for characters that are not allowed in XML
xml_replacement = dict( [ (eval('"\\x' + ('0'+hex(i)[2:])[-2:] + '"'), '' )  for i in xrange( 32 )] )
del xml_replacement[ '\x09' ]
del xml_replacement[ '\x0a' ]
del xml_replacement[ '\x0d' ]


#------------------------------------------------------------------------------
class MinfXMLWriter( MinfWriter ):
  '''
  Specialization of L{MinfWriter} class for writing XML minf format.
  '''
  
  name = 'XML'

  def __init__( self, file, reducer, 
                encoding = 'utf-8',
                level = 0,
                append=False ):
    self.__file = file
    self.reducer = createMinfReducer( reducer )
    self.encoder = codecs.getencoder( encoding )
    self.level = level
    self.indentString = '  '
    if not append:
      self._writeLine( '<?xml version="1.0" encoding=' + \
                        xml_quoteattr( encoding ) + ' ?>' )
      self._encodeAndWriteLine( '<' + minfTag + ' ' + expanderAttribute + \
        '=' + xml_quoteattr( reducer ) + '>' )
  
  
  def close( self ):
    if self.__file is not None:
      self.__file.flush()
      self._encodeAndWriteLine( '</' + minfTag + '>' )
      self.__file = None
  
  
  def write( self, value ):
    minfNodeIterator = self.reducer.reduce( value )
    for minfNode in minfNodeIterator:
      self._write( minfNodeIterator, minfNode, 0, None )
      
  
  def _write( self, minfNodeIterator, minfNode, level, name ):
    if minfNode is Undefined:
      minfNode = minfNodeIterator.next()
    attributes = {}
    if name is not None:
      attributes[ nameAttribute ] = name
    if isinstance( minfNode, StartStructure ):
      if minfNode.type == listStructure:
        naming = False
        stringNaming = False
        length = minfNode.attributes.get( 'length' )
        if length:
          attributes[ lengthAttribute ] = length
        tag = listTag
      elif minfNode.type == dictStructure:
        naming = True
        stringNaming = False
        length = minfNode.attributes.get( 'length' )
        if length:
          attributes[ lengthAttribute ] = length
        tag = dictionaryTag
      else:
        naming = True
        stringNaming = True
        tag = factoryTag
        attributes[ objectTypeAttribute ] = minfNode.type
      if attributes:
        attributes = ' ' + ' '.join( [ n + '='+ xml_quoteattr( unicode( v ) ) for n, v in attributes.iteritems()] )
      else:
        attributes = ''
      self._encodeAndWriteLine( '<' + tag + attributes + '>', level )
      ntype = minfNode.type
      for minfNode in minfNodeIterator:
        if isinstance( minfNode, EndStructure ):
          if ntype != minfNode.type:
            raise MinfError( _( 'Wrong Minf structure ending, expecting %(exp)s instead of %(rcv)s' ) % \
              { 'exp': ntype, 'rcv': minfNode.type } )
          self._encodeAndWriteLine( '</' + tag + '>', level )
          break
        elif naming:
          if isinstance( minfNode, ( str, unicode ) ):
            self._write( minfNodeIterator, Undefined, level+1, minfNode )
          elif minfNode is None:
            if not stringNaming:
              self._write( minfNodeIterator, minfNode, level+1, None )
            self._write( minfNodeIterator, Undefined, level+1, None )
          else:
            self._write( minfNodeIterator, minfNode, level+1, None )
            self._write( minfNodeIterator, Undefined, level+1, None )
        else:
          self._write( minfNodeIterator, minfNode, level+1, None )
    elif isinstance( minfNode, EndStructure ):
      raise MinfError( _( 'Unexpected Minf structure ending: %s' ) % ( minfNode.type, ) )
      level -= 1
    else:
      if attributes:
        attributesXML = ' ' + ' '.join( [ n + '='+ xml_quoteattr( unicode( v ) ) for n, v in attributes.iteritems()] )
      else:
        attributesXML = ''
      if minfNode is None:
        self._encodeAndWriteLine( '<' + noneTag + attributesXML + '/>', level )
      elif isinstance( minfNode, bool ):
        if minfNode:
          self._encodeAndWriteLine( '<' + trueTag + attributesXML + '/>', level )
        else:
          self._encodeAndWriteLine( '<' + falseTag + attributesXML + '/>', level )
      elif isinstance( minfNode, ( int, float, long ) ):
        self._encodeAndWriteLine( '<'+ numberTag + attributesXML + '>' + unicode( minfNode ) + '</' + \
                        numberTag + '>', level )
      elif isinstance( minfNode, ( str, unicode ) ):
        
        if type(minfNode) is str:
          try:
            minfNode=minfNode.decode("utf-8")
          except UnicodeDecodeError:
            minfNode=minfNode.decode("iso-8859-1")
        self._encodeAndWriteLine( '<' + stringTag + attributesXML + '>' + \
          xml_escape( minfNode, xml_replacement ) + '</' + stringTag + '>', level )
      elif hasattr( minfNode, '__minfxml__' ):
        minfNode.__minfxml__( self, attributes, level )
      else:
        raise MinfError( _( 'Cannot save an object of type %s as an XML atom' ) % ( str(type(minfNode)), ) )
    
  
  def _encodeAndWriteLine( self, line, level=0 ):
    self._writeLine( self.encoder( line )[ 0 ], level=level )
    
  
  def _writeLine( self, line, level=0 ):
    if self.level is None:
      indent = ''
      nl = ''
      level = None
    else:
      indent = self.indentString * ( self.level + level )
      nl = '\n'
    self.__file.write( indent + line + nl )
  
  
  def flush( self ):
    self.__file.flush()
  
  
  def change_file( self, file ):
    self.__file = file



#! /usr/bin/env python
##########################################################################
# CASPER - Copyright (C) CEA, 2013
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

try:
    from traits.api import CTrait
except ImportError:
    from enthought.traits.api import CTrait


_type_to_trait_id = {
    int: 'Int',
    unicode: 'Unicode',
    str: 'Str',
    float: 'Float'
}


def trait_ids(trait):
    """Return the type of the trait: File, Enum etc...

    Parameters
    ----------
    trait: trait instance (mandatory)
        a trait instance

    Returns
    -------
    main_id: list
        the string description (type) of the input trait
    """
    main_id = trait.handler.__class__.__name__
    #print main_id
    inner_ids = []
    if main_id == 'TraitCoerceType':
        real_id = _type_to_trait_id.get(trait.handler.aType)
        if real_id:
            main_id = real_id
    elif main_id == 'TraitCompound':
        #print trait.handler, dir(trait.handler)
        #print trait.handler.handlers
        trait_description = []
        for sub_trait in trait.handler.handlers:
            sub_c_trait = CTrait(0)
            sub_c_trait.handler = sub_trait
            trait_description.extend(trait_ids(sub_c_trait))
        return trait_description
    else:
        inner_id = '_'.join((trait_ids(i)[0]
                             for i in trait.handler.inner_traits()))
        if not inner_id:
            klass = getattr(trait.handler, 'klass', None)
            if klass is not None:
                inner_ids = [i.__name__ for i in klass.__mro__]
            else:
                inner_ids = []
        else:
            inner_ids = [inner_id]
    if inner_ids:
        return [main_id + '_' + i for i in inner_ids]
    else:
        return [main_id]


# -*- coding: utf-8 -*-
#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

'''

@author: Dominique Geffroy
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''

from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.qt_gui.qt_backend import QtWebKit

class TextEditWithSearch(QtGui.QTextEdit):
  """
  A QTextEdit with search feature to search a piece of text in the QTextEdit content.
  """
  def __init__(self, *args):
    super(TextEditWithSearch, self).__init__(*args)
    self.searchText=""
  
  def keyPressEvent(self, keyEvent):
    if (keyEvent.matches(QtGui.QKeySequence.Copy)):
      self.copy()
    elif (keyEvent.matches(QtGui.QKeySequence.SelectAll)):
      self.selectAll()
    elif (keyEvent.matches(QtGui.QKeySequence.Find)):
      self.search()
    elif (keyEvent.matches(QtGui.QKeySequence.FindNext) ):
      self.searchNext()
    else:
      QtGui.QAbstractScrollArea.keyPressEvent(self, keyEvent)
       
  def customMenu(self):
    menu=self.createStandardContextMenu()
    menu.addSeparator()
    menu.addAction("Find", self.search, QtGui.QKeySequence.Find) # Key_Control
    menu.addAction("Find next", self.searchNext, QtGui.QKeySequence.FindNext)
    return menu
  
  def contextMenuEvent(self, event):
    menu=self.customMenu()
    menu.exec_(event.globalPos())
  
  def search(self):
    (res, ok)=QtGui.QInputDialog.getText(self, "Find", "Text to find :", QtGui.QLineEdit.Normal, self.searchText)
    if ok:
      self.searchText=res
    if self.searchText and ok:
      self.moveCursor(QtGui.QTextCursor.Start)
      self.find(self.searchText) # not case sensitive, not whole word, forward
      
  def searchNext(self):
    if self.searchText:
      self.find(self.searchText) # not case sensitive, not whole word


class TextBrowserWithSearch(QtGui.QTextBrowser):
  """
  A QTextBrowser with search feature to search a piece of text in the QTextBrowser content.
  """
  def __init__(self, *args):
    super(TextBrowserWithSearch, self).__init__(*args)
    self.searchText=""

  def keyPressEvent(self, keyEvent):
    if (keyEvent.matches(QtGui.QKeySequence.Find)):
      self.search()
    elif (keyEvent.matches(QtGui.QKeySequence.FindNext) ):
      self.searchNext()
    elif (keyEvent.matches(QtGui.QKeySequence.Refresh) ):
      self.reload()
    else:
      QtGui.QTextBrowser.keyPressEvent(self, keyEvent)
  
  def customMenu(self):
    menu=self.createStandardContextMenu()
    menu.addSeparator()
    menu.addAction("Find", self.search, QtGui.QKeySequence.Find ) # Key_Control
    menu.addAction("Find next", self.searchNext, QtGui.QKeySequence.FindNext)
    menu.addAction("Reload", self.reload, QtGui.QKeySequence.Refresh)
    return menu
  
  def contextMenuEvent(self, event):
    menu=self.customMenu()
    menu.exec_(event.globalPos());
  
  def search(self):
    (res, ok)=QtGui.QInputDialog.getText(self, "Find", "Text to find :", QtGui.QLineEdit.Normal, self.searchText)
    if ok:
      self.searchText=res
    if self.searchText and ok:
      self.moveCursor(QtGui.QTextCursor.Start)
      self.find(self.searchText) # not case sensitive, not whole word, forward
      
  def searchNext(self):
    if self.searchText:
      self.find(self.searchText) # not case sensitive, not whole word

  def setSource(self, textUrl):
    """
    @type textUrl : string
    @param textUrl : l'url du fichier source
    """
    QtGui.QTextBrowser.setSource(self, QtCore.QUrl(textUrl))


class WebBrowserWithSearch(QtWebKit.QWebView):
  """
  A QWebView with search feature to search a piece of text in the QWebView content.
  """
  def __init__(self, *args):
    super(QtWebKit.QWebView, self).__init__(*args)
    self.searchText=""
    self.findAction = QtGui.QAction( _t_( 'Find' ), self )
    self.findAction.setShortcut( QtGui.QKeySequence.Find )
    self.connect( self.findAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.search )
    self.findNextAction = QtGui.QAction( _t_( 'Find next' ), self )
    self.findNextAction.setShortcut( QtGui.QKeySequence.FindNext )
    self.connect( self.findNextAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.searchNext )
    self.findPreviousAction = QtGui.QAction( _t_( 'Find previous' ), self )
    self.findPreviousAction.setShortcut( QtGui.QKeySequence.FindPrevious )
    self.connect( self.findPreviousAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.searchPrevious )
    self.zoomInAction = QtGui.QAction( _t_( 'Zoom in' ), self )
    self.zoomInAction.setShortcut( QtCore.Qt.CTRL + QtCore.Qt.Key_Plus )
    self.connect( self.zoomInAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.zoomIn )
    self.zoomOutAction = QtGui.QAction( _t_( 'Zoom out' ), self )
    self.zoomOutAction.setShortcut( QtCore.Qt.CTRL + QtCore.Qt.Key_Minus )
    self.connect( self.zoomOutAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.zoomOut )
    self.zoomOneAction = QtGui.QAction( _t_( 'Reset zoom' ), self )
    self.zoomOneAction.setShortcut( QtCore.Qt.CTRL + QtCore.Qt.Key_Equal )
    self.connect( self.zoomOneAction, QtCore.SIGNAL( 'triggered(bool)' ),
      self.zoomOne )
    self.addAction( self.findAction )
    self.addAction( self.findNextAction )
    self.addAction( self.findPreviousAction )
    self.addAction( self.zoomInAction )
    self.addAction( self.zoomOutAction )
    self.addAction( self.zoomOneAction )

  def customMenu(self):
    menu=QtGui.QMenu(self)
    menu.addAction( self.pageAction( QtWebKit.QWebPage.Back ) )
    menu.addAction( self.pageAction( QtWebKit.QWebPage.Forward ) )
    ra = self.pageAction( QtWebKit.QWebPage.Reload )
    if ra.shortcut() != QtGui.QKeySequence.Refresh:
      ra.setShortcut( QtGui.QKeySequence.Refresh )
    menu.addAction( ra )
    menu.addAction( self.pageAction( QtWebKit.QWebPage.Stop ) )
    menu.addSeparator()
    menu.addAction( self.findAction )
    menu.addAction( self.findNextAction )
    menu.addAction( self.findPreviousAction )
    menu.addSeparator()
    menu.addAction( self.zoomInAction )
    menu.addAction( self.zoomOutAction )
    menu.addAction( self.zoomOneAction )
    return menu

  def contextMenuEvent(self, event):
    menu=self.customMenu()
    menu.exec_(event.globalPos());

  def search(self, void):
    text = self.selectedText()
    if not text:
      text = self.searchText
    (res, ok)=QtGui.QInputDialog.getText(self, _t_( "Find" ), _t_( "Text to find :" ), QtGui.QLineEdit.Normal, text )
    if ok:
      self.searchText=res
    if self.searchText and ok:
      self.findText( self.searchText )

  def searchNext(self, void):
    if self.searchText:
      # not case sensitive, not whole word
      self.findText(self.searchText, QtWebKit.QWebPage.FindWrapsAroundDocument )

  def searchPrevious(self, void):
    if self.searchText:
      self.findText(self.searchText,
        QtWebKit.QWebPage.FindFlags( QtWebKit.QWebPage.FindBackward \
        + QtWebKit.QWebPage.FindWrapsAroundDocument ) )

  def setSource(self, textUrl):
    self.load(QtCore.QUrl(textUrl))

  def zoomIn( self, void ):
    self.setZoomFactor( self.zoomFactor() * 1.1 )
    #self.setTextSizeMultiplier( self.textSizeMultiplier() * 1.1 )

  def zoomOut( self, void ):
    self.setZoomFactor( self.zoomFactor() / 1.1 )
    #self.setTextSizeMultiplier( self.textSizeMultiplier() / 1.1 )

  def zoomOne( self, void ):
    self.setZoomFactor( 1. )
    self.setTextSizeMultiplier( 1. )





# -*- coding: iso-8859-1 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.


'''
Base classes for reading various minf formats (XML, HDF5, Python's pickle, etc.)

@author: Yann Cointepas
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''

__docformat__ = "epytext en"


from soma.translation import translate as _

#------------------------------------------------------------------------------
class RegisterMinfReaderClass( type ):
  '''
  RegisterMinfReaderClass is used as metaclass of L{MinfReader} to automatically
  register all classes derived from L{MinfReader}.
  '''
  def __init__( cls, name, bases, dict ):
    if cls.name is not None:
      MinfReader._allReaderClasses[ cls.name ] = cls


#------------------------------------------------------------------------------
class MinfReader( object ):
  '''
  Class derived from MinfReader are responsible of reading a specific format of
  minf file. This version only support XML format but other formats may be added
  later (such as HDF5).
  All classes derived from L{MinfReader} must define a L{name} class attribute
  and L{nodeIterator} method. The constructor of L{MinfReader} derived class
  must be callable without arguments (except C{self}), it does not have to be
  overloaded.
  '''
  __metaclass__ = RegisterMinfReaderClass
  
  #: all classes derived from L{MinfReader} are automatically stored in that
  #: dictionary (keys are formats name and values are class objects).
  _allReaderClasses = {}

  #: class derived from L{MinfReader} must set a format name in this attribute.
  name = None
  
  
  
  def createReader( format ):
    '''
    This static method create a L{MinfReader} instance by looking for a registered
    L{MinfReader} derived classes named C{format}.
    @param format: name of the minf format.
    @type  format: string
    @returns: L{MinfReader} derived class instance.
    '''
    reader = MinfReader._allReaderClasses.get( format )
    if reader is None:
      raise( _( 'No minf reader for format "%(format)s", possible formats are: %(possible)s' ) % \
             { 'format': format, 'possible': ', '.join( ['"' + i + '"' for i in MinfReader._allReaderClasses ] ) } )
    return reader()
  createReader = staticmethod( createReader )


  def reduction( self, source ):
    '''
    Return a pair C{( reduction, buffer )} where C{reduction} is the name of
    the L{MinfReducer<soma.minf.tree.MinfReducer} stored in the minf file and
    C{buffer} is the data that have been read from file.
    '''
  def nodeIterator( self, source ):
    '''
    Return an iterator on all the nodes of the minf tree corresponding to the content
    of the C{source} minf file.
    @param source: source of the minf file.
    @type  source: string of file object
    '''



import os, re
from glob import glob
from soma.qt_gui import qt_backend
if __name__ == '__main__':
    qt_backend.set_qt_backend('PyQt4')
from soma.qt_gui.qt_backend.QtGui import QDialog, QTableWidgetItem, QVBoxLayout
from soma.qt_gui.qt_backend import loadUi


class CATIScanSelector(QDialog):
    '''
    Dirty implementation of selection dialog for CATI FOM.
    This class will be modified or removed and should not be used directly.
    '''
    labels = ['Study', 'Subject', 'Time point']
    glob = ('*', 'CONVERTED', '*', '*', '*')
    path_re = ('(?P<study>.*)', '(?P<test>CONVERTED)', '(?P<center>.*)',
        '(?P<subject>.*)', '(?P<time_point>.*)')
    match_groups = ['study', 'subject', 'time_point']

    def __init__(self, directory=None, parent=None):
        super(CATIScanSelector, self).__init__(parent=parent)
        ui = os.path.join(__file__[: __file__.rfind('.')] + '.ui')
        ui_obj = loadUi(ui, self)
        if qt_backend.get_qt_backend() == 'PySide':
            self.table = ui_obj.table
            self.status = ui_obj.status
            layout = QVBoxLayout(self)
            self.setLayout(layout)
            layout.addWidget(ui_obj)
        if directory:
            self.reset(directory)


    def reset(self, directory):
        '''
        Clear the selection dialog and parse directory to extract selectable
        attributes.
        '''
        self.table.clear()
        self.table.setColumnCount(3)
        self.table.setHorizontalHeaderLabels(self.labels)
        self.table.setRowCount(0)
        studies = set()
        time_points = set()
        subjects = set()
        path_re = re.compile(os.path.join(os.path.join(
            directory, *self.path_re)))
        self.attributes = []
        for i in glob(os.path.join(directory, *self.glob)):
            match = path_re.match(i)
            if match:
                self.attributes.append(match.groupdict())
                row = self.table.rowCount()
                self.table.setRowCount(row + 1)
                for i in xrange(len(self.match_groups)):
                    self.table.setItem(row, i,
                        QTableWidgetItem(match.group(self.match_groups[i])))
        self.table.resizeColumnsToContents()
        self.table.itemSelectionChanged.connect(self._update_selection)
        self._update_selection()


    def _update_selection(self):
      '''
      Called each time the selection is changed to update QDialog (e.g. the
      status bar with number of items selected).
      '''
      count = 0
      for range in self.table.selectedRanges():
        count += range.bottomRow() - range.topRow() + 1
      if count:
        self.status.setText('%d item selected' % count)
      else:
        self.status.setText('No item selected')


    def select(self):
        '''
        After a modal execution of the QDialog, returns a list of dictionaries
        (one for each selected item) or False if the user did not pressed "Ok".
        '''
        if self.exec_() == self.Accepted:
            result = []
            for range in self.table.selectedRanges():
                result.extend(
                    self.attributes[range.topRow() : range.bottomRow() + 1])
            return result
        else:
            return False


class BrainVISAScanSelector(CATIScanSelector):
    '''
    Dirty implementation of selection dialog for Morphologist FOM.
    This class will be modified or removed and should not be used directly.
    '''
    labels = ['Center', 'Subject', 'Acquisition']
    glob = ('*', '*', 't1mri', '*')
    path_re = ('(?P<center>.*)', '(?P<subject>.*)', 't1mri',
        '(?P<acquisition>.*)')
    match_groups = ['center', 'subject', 'acquisition']


class FileAttributeSelection(object):
    '''
    A FileAttributeSelection is used to select multiple files according to
    their attributes as defined in a given FOM.

    Example:

    from soma.gui.file_selection import FileAttributeSelection
    file_selection = FileAttributeSelection()
    selection = file_selection.select(fom, 'unused', ['unused'], sys.argv[1])

    '''
    _selectors = {
        ('.*cati.*', '.*'): CATIScanSelector,
        ('.*brainvisa.*', '.*'): BrainVISAScanSelector,
        ('morphologist.*', '.*'): BrainVISAScanSelector,
    }

    def __init__(self):
        self._selector_rules = [(re.compile(i[0]), re.compile(i[1]), j) \
            for i,j in self._selectors.iteritems()]


    def find_selector(self, fom_name, process_name, parameters):
        '''
        Find the selector class allowing to select multiple files for a given
        process.
        fom_name : FileOrganisationModel name
        process_name: name of the process
        parameters : list of parameter names

        return: a class derived from QDialog or None if none is found for the
                given parameters.

        Actual implementation is incomplete. It ignores parameters and only
        use a pattern matching on the FOM name and process_name.
        '''
        for fom_re, process_re, selector_class in self._selector_rules:
            if fom_re.match(fom_name) and process_re.match(process_name):
                return selector_class
        return None


    def select(self,fom_name, process, parameters, directory, parent=None):
        '''
        Find a selector class according to fom, process_name and parameters
        and display the selecion dialog.
        Return value can be:
            - None if no selector class had been found
            - False if the used did not closed the dialog by clicking "Ok"
            - a list of dictionaries (one for each selected item)
        '''
        selector_class = self.find_selector(fom_name, process, parameters)
        print 'selection_class', selector_class
        if selector_class:
            return selector_class(directory=directory).select()
        return None



if __name__ == '__main__':
    import sys
    from soma.qt_gui.qt_backend.QtGui import QApplication
    from pprint import pprint

    from soma.application import Application
    # First thing to do is to create an Application with name and version
    app = Application('soma.fom', '1.0')
    # Register module to load and call functions before and/or after
    # initialization
    app.plugin_modules.append('soma.fom')
    # Application initialization (e.g. configuration file may be read here)
    app.initialize()

    from pprint import pprint
    #fom = app.fom_manager.load_foms( 'morphologist-brainvisa-1.0' )
    fom_name = 'morphologist-brainvisa-1.0'
    #fom_name='morphologist-cati-memento'
    fom = app.fom_manager.load_foms(fom_name)


    app = QApplication(sys.argv)

    if len(sys.argv) >= 2:
        direct = sys.argv[1]
    else:
        direct = '/home/mb236582/datafom'
        #direct = '/nfs/neurospin/cati/cati_shared'
    file_selection = FileAttributeSelection()
    selection = file_selection.select(fom_name, 'unused', ['unused'], direct)

    #pprint( selection )




# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

'''
@author: Yann Cointepas
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
'''
__docformat__ = "restructuredtext en"


from soma.qt_gui.qt_backend.QtCore import Qt, QAbstractTableModel, SIGNAL, QModelIndex
from soma.qt_gui.qt_backend.QtGui import QBrush
from soma.qt_gui.qt_backend import QtCore, get_qt_backend
use_qvariant = False
if get_qt_backend() == 'PyQt4':
  try:
    import sip
    use_qvariant = sip.getapi( 'QVariant' ) < 2
    if use_qvariant:
      from PyQt4.QtCore import QVariant
  except:
    pass # PySide doesn't have QVariant at all.
import operator

class SimpleTable( QAbstractTableModel ):
  '''
  A simple read-only table model allowing to display data with L{PyQt4.QtGui.QTableView}
  
  Example
  -------
    from PyQt4.uic import loadUi
    from PyQt4.QtGui import QWidget
    from PyQt4.QtCore import SIGNAL
    from soma.qt4gui.api import SimpleTable

    def clicked( index ):
      global m
      print 'Selected', m.row( index.row() )[ index.column() ], 'in', m.row( index.row() )

    m = SimpleTable( ( 'a', 'b', 'c' ), ( ( 1, 2, 3 ), ( 4, 5, 6 ) ) )
    w = QWidget()
    loadUi( '/tmp/test.ui', w )
    w.tableView.setModel( m )
    w.tableView.connect( w.tableView, SIGNAL( 'clicked ( const QModelIndex & )' ), clicked )
    w.show()
  '''
  def __init__(self, header, data=(), parent=None ):
    '''header = series of unicode
    '''
    QAbstractTableModel.__init__( self, parent ) 
    self._header = tuple( header )
    self._data = []
    self._background = {}
    for row in data:
      self.addRow( row )
      
  def addRow( self, row ):
    parent = QModelIndex()
    index = len( self._data )
    self.beginInsertRows( parent, index, index ) 
    self._data.append( tuple( row ) + ( index, ) )
    self.endInsertRows()

  def clear( self ):
    parent = QModelIndex()
    last = len( self._data ) - 1
    self.beginRemoveRows( parent, 0, last ) 
    # Remove all elements from self._data
    self._data[:] = []
    self._background = {}
    self.endRemoveRows()
  
  def setRow( self, index, row ):
    self._data[ index ] = tuple( row ) + ( self._data[ index ][ -1 ], )
    self.emit( SIGNAL( 'dataChanged ( const QModelIndex &, const QModelIndex & )' ), self.createIndex( index, 0 ), self.createIndex( index, len( row )-1 ) )

  def row( self, index ):
    return self._data[ index ][ :-1 ]
  
  
  def rowCount( self, parent ): 
    return len( self._data ) 
 
  
  def columnCount( self, parent ): 
    return len( self._header ) 

  def setRowBackgroundColor( self, index, color ):
    if color is None:
      self._background.pop( index, None ) 
    else:
      self._background[ index ] = color 
    self.emit( SIGNAL( 'dataChanged ( const QModelIndex &, const QModelIndex & )' ), self.createIndex( index, 0 ), self.createIndex( index, len( self._data[ index ] )-2 ) )
  
  def data( self, index, role=Qt.DisplayRole ):
    if use_qvariant:
      result = QVariant()
    else:
      result = None
    if index.isValid():
      if role == Qt.DisplayRole: 
        value = self._data[ index.row() ][ index.column() ]
        if value is not None:
          if use_qvariant:
            result = QVariant( value )
          else:
            result = value
      elif role == Qt.BackgroundRole:
        background = self._background.get( index.row() )
        if background is not None:
          if use_qvariant:
            result = QVariant( QBrush( background ) )
          else:
            result = QBrush( background )
    return result
  
  def headerData( self, col, orientation, role ):
    if orientation == Qt.Horizontal and role == Qt.DisplayRole:
      if use_qvariant:
        return QVariant( self._header[ col ] )
      else:
        return self._header[ col ]
    if use_qvariant:
      return QVariant()
    else:
      return None

  
  def sort( self, column, order ):
    """Sort table by given column number.
    """
    self.emit( SIGNAL( "layoutAboutToBeChanged()" ) )
    self._data.sort( key=operator.itemgetter( column ), reverse=(order == Qt.DescendingOrder) )
    self.emit( SIGNAL( "layoutChanged()" ) )

  def sortedIndex( self, index ):
    return self._data[ index ][ -1 ]
  



#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the 
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

"""
This module enables to make a function to be executed in qt thread (main thread). 
It is useful when you want to call qt functions from another thread.
It enables to do thread safe calls because all tasks sent are executed in the same thread (qt main thread).

@author: Dominique Geffroy
@organization: U{NeuroSpin<http://www.neurospin.org>} and U{IFR 49<http://www.ifr49.org>}
@license: U{CeCILL version 2<http://www.cecill.info/licences/Licence_CeCILL_V2-en.html>}
"""

import sys
import threading
from soma.qt_gui.qt_backend.QtCore import QObject, QTimer, QEvent, SIGNAL, QCoreApplication
from soma import singleton

class FakeQtThreadCall( QObject ):
  '''
  Fake L{QtThreadCall} that behave as if always used from main thread.
  '''
  def isInMainThread():
    return True
  isInMainThread = staticmethod( isInMainThread )
  
  
  def push( function, *args, **kwargs ):
    if kwargs is None or not kwargs:
      function( *args )
    else:
      function( *args, **kwargs )
  push = staticmethod( push )

  def call( function, *args, **kwargs ):
    if kwargs is None or not kwargs:
      return function( *args )
    else:
      return function( *args, **kwargs )
  call = staticmethod( call )



class QtThreadCall( singleton.Singleton, QObject ):
  """
  This object enables to send tasks to be executed by qt thread (main thread). 
  This object must be initialized in qt thread. 
  It starts a QTimer and periodically execute tasks in its actions list.
  
  @type lock: RLock
  @ivar lock: lock to prevent concurrent access to actions list
  @type actions: list
  @ivar actions: tasks to execute
  @type mainThread: Thread
  @ivar mainThread: current thread at object initialisation
  @type timer: QTimer
  @ivar timer: timer to wake this object periodically
  """
  def __singleton_init__( self ):
    # QObject.__init__( self, None )
    super( QtThreadCall, self ).__singleton_init__( None )
    self.lock = threading.RLock()
    self.actions = []
    # look for the main thread
    mainthreadfound = False
    for thread in threading.enumerate():
      if isinstance( thread, threading._MainThread ):
        mainthreadfound = True
        self.mainThread=thread
        break
    if not mainthreadfound:
      print 'Warning: main thread not found'
      self.mainThread=threading.currentThread()

  def _postEvent( self ):
    class QtThreadCallEvent( QEvent ):
      def __init__( self, qthreadcall ):
        QEvent.__init__( self, QEvent.Type( QEvent.User + 24 ) )
        self.qthreadcall = qthreadcall
    QCoreApplication.instance().postEvent( self, 
      QtThreadCallEvent( self ) )

  def event( self, e ):
    sys.stdout.flush()
    if e.type() == QEvent.User + 24:
      self.doAction()
      return True
    return QObject.event( self, e )

  def push( self, function, *args, **kwargs ):
    """
    Add a function call to the actions list. the call is executed immediatly if current thread is main thread.
    
    @type function: function
    @param function: the function to call in main thread.
    """
    if self.isInMainThread():
      if kwargs is None or len( kwargs ) == 0:
        apply( function, args )
      else:
        apply( function, args, kwargs )
    else:
      self.lock.acquire()
      try:
        self.actions.append( ( function, args, kwargs ) )
        self._postEvent()
      finally:
        self.lock.release()

  def call( self, function, *args, **kwargs ):
    """
    Send the function call to be executed in the qt main thread and wait for the result.
    
    @type function: function
    @param function: the function to call in main thread.
     
    @return: function call's result
    """
    if self.isInMainThread():
      if kwargs is None or len( kwargs ) == 0:
        return apply( function, args )
      return apply( function, args, kwargs )
    else:
      semaphore = threading.Semaphore( 0 )
      self.lock.acquire()
      try:
        self.actions.append( ( self._callAndWakeUp, ( semaphore, function, args, kwargs ), {} ) )
        self._postEvent()
      finally:
        self.lock.release()
      semaphore.acquire() # block until semaphore is released in _callAndWakeUp method
      result = semaphore._mainThreadActionResult
      exception = semaphore._mainThreadActionException
      if exception is not None:
        e, v, t = exception
        raise e, v, t
      return result

  def _callAndWakeUp( self, semaphore, function, args, kwargs ):
    """
    Call the function, set the result in semaphore attributes and release the semaphore.
    
    @type semaphore: threading.Semaphore
    @ivar semaphore: thread which has added this task is blocked on this semaphore. function call's result will be kept in this semaphore attributes.
    @type function: function
    @param function: the function to call in main thread.
    """
    semaphore._mainThreadActionResult = None
    semaphore._mainThreadActionException = None
    try:
      if kwargs is None or len( kwargs ) == 0:
        semaphore._mainThreadActionResult = function( *args )
      else:
        semaphore._mainThreadActionResult = function( *args, **kwargs )
    except:
      semaphore._mainThreadActionException = sys.exc_info()
    semaphore.release() # release the semaphore to unblock the thread which waits for the function call's result

  def isInMainThread( self ):
    """
    @rtype: boolean
    @return: True if current thread is main thread
    """ 
    #return threading.currentThread().getName() == 'MainThread'
    return threading.currentThread() is self.mainThread

  def doAction( self ):
    """
    This method is called each time the timer timeout.
    It executes all functions in actions list.
    """
    self.lock.acquire()
    try:
      actions = self.actions
      #print "actions to do", self.actions
      self.actions = []
    finally:
      self.lock.release()
    for ( function, args, kwargs ) in actions:
      try:
        if kwargs is None or len( kwargs ) == 0:
          apply( function, args )
        else:
          apply( function, args, kwargs )
      except:
        # Should call a customizable function here
        raise



# -*- coding: utf-8 -*-
import sipconfig
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.controller import trait_ids, Controller
from soma.utils.functiontools import partial, SomaPartial
from soma.qt4gui.api import TimeredQLineEdit
from soma.global_naming import GlobalNaming
from soma.gui.icon_factory import IconFactory
try:
    from traits.api import File, Directory
except ImportError:
    from enthought.traits.api import File, Directory



#-------------------------------------------------------------------------------
class StrCreateWidget( object ):
  @staticmethod
  def is_valid_trait( trait ):
    return True
  
  @staticmethod
  def create_widget( parent, name, trait, value ):
    attribute_widget = TimeredQLineEdit( parent )
    label = trait.label
    if not label:
      label = name
    if label is not None:
      label_widget = QtGui.QLabel( label, parent )
    else:
      label_widget = None
    
    return ( attribute_widget, label_widget )
    

  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    """GUI modified so update traits"""
    if isinstance( attribute_widget, TimeredQLineEdit ):
      try:
        setattr(controller_widget.controller, name, unicode( attribute_widget.text() ) )
      except:
        print 'Warning - StrCreateWidget.update_controller: controller widget set failure'
    else: 
      setattr(controller_widget.controller, name, unicode( attribute_widget.text_widget.text() ) )
    

  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):
    """Traits modified so update GUI""" 
    if isinstance( attribute_widget, TimeredQLineEdit ):
      attribute_widget.setText( unicode( getattr( controller_widget.controller, name, '' ) ) )
    else:   
      attribute_widget.text_widget.setText( unicode( getattr( controller_widget.controller, name, '' ) ) )
  

  @classmethod
  def connect_controller( cls, controller_widget, name, attribute_widget, label_widget ):  
    widget_hook = partial( cls.update_controller, controller_widget, name, attribute_widget )  
    if isinstance( attribute_widget, TimeredQLineEdit ):
      attribute_widget.connect( attribute_widget, QtCore.SIGNAL( 'userModification' ), 
                                widget_hook )
    else:
      attribute_widget.connect( attribute_widget.text_widget, QtCore.SIGNAL( 'userModification' ), 
                                widget_hook )  
                                            
    controller_hook = SomaPartial( cls.update_controller_widget, controller_widget, name, attribute_widget, label_widget )
    controller_widget.controller.on_trait_change( controller_hook, name=name )
    attribute_widget._controller_connections = ( widget_hook, controller_hook )

  @staticmethod
  def disconnect_controller( controller_widget, name, attribute_widget, label_widget ):
    widget_hook, controller_hook = attribute_widget._controller_connections
    controller_widget.controller.on_trait_change( controller_hook, name=name, remove=True )
    if isinstance( attribute_widget, TimeredQLineEdit ):
      attribute_widget.disconnect( attribute_widget, QtCore.SIGNAL( 'userModification' ), 
                                   widget_hook )
    else:
      attribute_widget.disconnect( attribute_widget.text_widget, QtCore.SIGNAL( 'userModification' ), 
                                   widget_hook )  
    del attribute_widget._controller_connections
    
  


#-------------------------------------------------------------------------------
class FloatCreateWidget( StrCreateWidget ):
  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    setattr( controller_widget.controller, name, float( attribute_widget.text() ) )
        

  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):
    attribute_widget.setText( unicode( getattr( controller_widget.controller, name, 0 ) ) )


    

#-------------------------------------------------------------------------------
class IntCreateWidget( FloatCreateWidget ):
  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    setattr( controller_widget.controller, name, int( attribute_widget.text() ) )


    

#-------------------------------------------------------------------------------
class LongCreateWidget( FloatCreateWidget ):
  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    setattr( controller_widget.controller, name, long( attribute_widget.text() ) )

    

#-------------------------------------------------------------------------------
class BoolCreateWidget( object ):
  @staticmethod
  def is_valid_trait( trait ):
    return True
  
  @staticmethod
  def create_widget( parent, name, trait, value ):
    attribute_widget = QtGui.QCheckBox( parent )
    label = trait.label
    if not label:
      label = name
    if label is not None:
      label_widget = QtGui.QLabel( label, parent )
    else:
      label_widget = None
    return ( attribute_widget, label_widget )


  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    setattr( controller_widget.controller, name, bool( attribute_widget.isChecked () ) )


  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):
    attribute_widget.setChecked( getattr( controller_widget.controller, name, False ) )


  @classmethod
  def connect_controller( cls, controller_widget, name, attribute_widget, label_widget ):
    widget_hook = partial( cls.update_controller, controller_widget, name, attribute_widget )
    attribute_widget.connect( attribute_widget, QtCore.SIGNAL( 'clicked()' ), 
                              widget_hook )
    controller_hook = SomaPartial( cls.update_controller_widget, controller_widget, name, attribute_widget, label_widget )
    controller_widget.controller.on_trait_change( controller_hook, name=name )
    attribute_widget._controller_connections = ( widget_hook, controller_hook )
    
    
  @staticmethod
  def disconnect_controller( controller_widget, name, attribute_widget, label_widget ):
    widget_hook, controller_hook = attribute_widget._controller_connections
    controller_widget.controller.on_trait_change( controller_hook, name=name, remove=True )
    attribute_widget.disconnect( attribute_widget, QtCore.SIGNAL( 'clicked()' ), 
                                 widget_hook )
    del attribute_widget._controller_connections
    



#-------------------------------------------------------------------------------
class EnumCreateWidget( object ):
  @staticmethod
  def is_valid_trait( trait ):
    return True
  
  @staticmethod
  def create_widget( parent, name, trait, value ):
    attribute_widget = QtGui.QComboBox( parent )
    values = trait.handler.values
    labels = getattr( trait, 'labels' )
    if not labels:
      labels = ()
    labels += ( None, ) * ( len( values ) - len( labels ) )
    labels = [ ( unicode( value ) if label is None else label ) for value, label in zip( values, labels ) ]
    attribute_widget._values = values
    for label in labels:
      attribute_widget.addItem( label )
    label = trait.label
    if not label:
      label = name
    if label is not None:
      label_widget = QtGui.QLabel( label, parent )
    else:
      label_widget = None
    return ( attribute_widget, label_widget )

  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    #default_value_change(getattr(controller_widget.controller,name),attribute_widget._values[ attribute_widget.currentIndex() ],attribute_widget.btn_default_value)
    setattr( controller_widget.controller, name, attribute_widget._values[ attribute_widget.currentIndex() ] )

  
  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):
    label = getattr( controller_widget.controller, name, None )
    if label is not None:
      attribute_widget.setCurrentIndex( attribute_widget._values.index( label ) )
  
  @classmethod
  def connect_controller( cls, controller_widget, name, attribute_widget, label_widget ):
    widget_hook = partial( cls.update_controller, controller_widget, name, attribute_widget )
    attribute_widget.connect( attribute_widget, QtCore.SIGNAL( 'activated( int )' ), 
                              widget_hook )
    controller_hook = SomaPartial( cls.update_controller_widget, controller_widget, name, attribute_widget, label_widget )
    controller_widget.controller.on_trait_change( controller_hook, name=name )
    attribute_widget._controller_connections = ( widget_hook, controller_hook )

    
  @staticmethod
  def disconnect_controller( controller_widget, name, attribute_widget, label_widget ):
    widget_hook, controller_hook = attribute_widget._controller_connections
    controller_widget.controller.on_trait_change( controller_hook, name=name, remove=True )
    attribute_widget.disconnect( attribute_widget, QtCore.SIGNAL( 'activated( int )' ), 
                              widget_hook )
    del attribute_widget._controller_connections

    

#-------------------------------------------------------------------------------
class StrEnumCreateWidget( object ):
  @staticmethod
  def is_valid_trait( trait ):
    return trait.values is not None
  
  @staticmethod
  def create_widget( parent, name, trait, value ):
    values = list( trait.values )
    if trait.labels:
      raise ValueError( 'labels is not allowed for Str with values' )
    attribute_widget = QtGui.QComboBox( parent )
    attribute_widget.setEditable( True )
    attribute_widget._values = values
    for label in values:
      attribute_widget.addItem( label )
    label = trait.label
    if not label:
      label = name
    if label is not None:
      label_widget = QtGui.QLabel( label, parent )
    else:
      label_widget = None
    return ( attribute_widget, label_widget )

  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    setattr( controller_widget.controller, name, unicode( attribute_widget.currentText() ) )
  
  
  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):
    value = getattr( controller_widget.controller, name, '' )
    try:
      index = attribute_widget._values.index( value )
    except ValueError:
      index = -1
    if index != -1:
      attribute_widget.setCurrentIndex( index )
    else:
      attribute_widget.setEditText( value )
  
  @classmethod
  def connect_controller( cls, controller_widget, name, attribute_widget, label_widget ):
    widget_hook = partial( cls.update_controller, controller_widget, name, attribute_widget )
    attribute_widget.connect( attribute_widget, QtCore.SIGNAL( 'editTextChanged ( const QString & )' ), 
                              widget_hook )
    controller_hook = SomaPartial( cls.update_controller_widget, controller_widget, name, attribute_widget, label_widget )
    controller_widget.controller.on_trait_change( controller_hook, name=name )
    attribute_widget._controller_connections = ( widget_hook, controller_hook )

    
  @staticmethod
  def disconnect_controller( controller_widget, name, attribute_widget, label_widget ):
    widget_hook, controller_hook = attribute_widget._controller_connections
    controller_widget.controller.on_trait_change( controller_hook, name=name, remove=True )
    attribute_widget.disconnect( attribute_widget, QtCore.SIGNAL( 'editTextChanged ( const QString & )' ), 
                                 widget_hook )
    del attribute_widget._controller_connections


#-------------------------------------------------------------------------------
class FileCreateWidget( object ):
  @staticmethod
  def is_valid_trait( trait ):
    return True
  
  @classmethod
  def create_widget( cls, parent, name, trait, value ):
    text_widget, label_widget = StrCreateWidget.create_widget( parent, name, trait, value )   
    attribute_widget = QtGui.QWidget( parent )
    horizontal_layout = QtGui.QHBoxLayout( attribute_widget )
    horizontal_layout.setContentsMargins( 0, 0, 0, 0 )
    horizontal_layout.addWidget( text_widget )
    attribute_widget.text_widget = text_widget   

    btn_browse = QtGui.QPushButton( attribute_widget )
    if trait.output is True:
        btn_browse.setIcon( QtGui.QIcon( IconFactory()._imageBrowseOutput) )
    if trait.output is False:
        btn_browse.setIcon( QtGui.QIcon( IconFactory()._imageBrowseInput) )
    if trait(name).is_trait_type( Directory):
      btn_browse.setIcon( QtGui.QIcon( IconFactory()._imageBrowseDir) )
    horizontal_layout.addWidget( btn_browse )
    btn_browse.setFixedSize( 20, 20 )
    attribute_widget.btn_browse = btn_browse
    attribute_widget._browse_hook = partial( cls.file_dialog, parent, name ,attribute_widget )
    attribute_widget.btn_browse.clicked.connect( attribute_widget._browse_hook )

    viewers = getattr( parent.controller, 'viewers', None)
    if viewers and name in viewers :
      btn_viewer = QtGui.QPushButton( attribute_widget )
      horizontal_layout.addWidget( btn_viewer )
      btn_viewer.setFixedSize( 20, 20 )
      btn_viewer.setIcon( QtGui.QIcon( IconFactory()._imageViewer ) )
      attribute_widget.btn_viewer = btn_viewer
      if hasattr( parent.controller, 'call_viewer' ):
        viewer_hook = partial( parent.controller.call_viewer, parent, name )
        attribute_widget._viewer_hook = viewer_hook
        attribute_widget.btn_viewer.clicked.connect( viewer_hook )
    return ( attribute_widget, label_widget )

  @staticmethod
  def file_dialog( parent, name,attribute_widget ):
    if hasattr( parent.controller.user_traits()[name], 'output' ) \
        and parent.controller.user_traits()[name].output:
      outputfile = True
    else:
      outputfile = False
    if sipconfig.Configuration().sip_version >= 0x040a00:
      #/nfs/neurospin/cati/cati_shared/MEMENTO/CONVERTED/001/0010020_LAFR/M000/MRI/3DT1/0010020_LAFR_M000_3DT1_S002.nii.gz
        #''
      if not outputfile:
        value = QtGui.QFileDialog.getOpenFileName ( parent, 'Select a file','/home/mb236582/datafom/subjects/subject02/t1mri/default_acquisition/subject02.nii', '', 
                                                    options=QtGui.QFileDialog.DontUseNativeDialog )
      else:
        value = QtGui.QFileDialog.getSaveFileName ( parent, 'Select a file','/home/mb236582/datafom/subjects/subject02/t1mri/default_acquisition/subject02.nii', '', 
                                                    options=QtGui.QFileDialog.DontUseNativeDialog )
    else:
      if outputfile:
        value = QtGui.QFileDialog.getOpenFileName ( self._widget, 'Select a file', '', '', 
                                                    0, QtGui.QFileDialog.DontUseNativeDialog )
      else:
        value = QtGui.QFileDialog.getSaveFileName ( self._widget, 'Select a file', '', '', 
                                                    0, QtGui.QFileDialog.DontUseNativeDialog )
    setattr( parent.controller, name, unicode( value ) )

    
  @staticmethod
  def update_viewer(controller_widget,name,attribute_widget):
    if hasattr(controller_widget.controller, 'viewers'):
      if name in controller_widget.controller.viewers:
        try:
          open(getattr(controller_widget.controller,name))
          attribute_widget.btn_viewer.setEnabled(True)
        except IOError:
          attribute_widget.btn_viewer.setEnabled(False)
    
      

  @staticmethod
  def update_controller( controller_widget, name, attribute_widget ):
    """GUI modified so update traits"""
    StrCreateWidget.update_controller( controller_widget, name, attribute_widget.text_widget )
    #StrCreateWidget.update_controller( controller_widget, name, attribute_widget)
  
  @staticmethod
  def update_controller_widget( controller_widget, name, attribute_widget, label_widget ):   
    """Traits modified so update GUI"""
    StrCreateWidget.update_controller_widget( controller_widget, name, attribute_widget.text_widget, label_widget )
    #StrCreateWidget.update_controller_widget( controller_widget, name, attribute_widget, label_widget )
   
    # To disable viewer if file doesn't exist
    #if controller_widget.controller.trait( name ).viewer is not None:
    if hasattr(controller_widget.controller, 'viewers'):
      if name in controller_widget.controller.viewers :
        try:
          open(getattr(controller_widget.controller,name))
          attribute_widget.btn_viewer.setEnabled(True)
        except IOError:
          attribute_widget.btn_viewer.setEnabled(False)
          
  
  @classmethod
  def connect_controller( cls, controller_widget, name, attribute_widget, label_widget ):
    #StrCreateWidget.connect_controller( controller_widget, name, attribute_widget.text_widget, label_widget )
    StrCreateWidget.connect_controller( controller_widget, name, attribute_widget.text_widget, label_widget )
    
    # Function call when traits change to update if viewer enabled or not
    controller_hook = SomaPartial( cls.update_viewer, controller_widget, name, attribute_widget )
    controller_widget.controller.on_trait_change( controller_hook, name )
    
    attribute_widget._controller_connections = controller_hook
    

  @staticmethod
  def disconnect_controller( controller_widget, name, attribute_widget, label_widget ):
    controller_hook = attribute_widget._controller_connections
    controller_widget.controller.on_trait_change( controller_hook, name, remove=True )
    StrCreateWidget.disconnect_controller( controller_widget, name, attribute_widget.text_widget, label_widget )
    del attribute_widget._controller_connections
    



#-------------------------------------------------------------------------------
class DirectoryCreateWidget( FileCreateWidget ):
  @staticmethod
  def file_dialog( controller_widget, name,attribute_widget ):
    value = QtGui.QFileDialog.getExistingDirectory ( controller_widget, 'Select a directory', '', 
       options=QtGui.QFileDialog.ShowDirsOnly | QtGui.QFileDialog.DontUseNativeDialog )
    setattr( controller_widget.controller, name, unicode( value ) )
    
    

def find_function_viewer(name_viewer):
  return GlobalNaming().get_object( name_viewer )
  
  
#def default_value_change(old,new,btn):
    #if old !=new:
        #btn.setVisible(True)




from __future__ import absolute_import

import logging

def create_logger( name ):
  logger = logging.getLogger()
  handler = logging.StreamHandler()
  formatter = logging.Formatter('%(levelname)s:%(funcName)s(%(lineno)s): %(message)s')
  handler.setFormatter( formatter )
  logger.addHandler( handler )
  logger.setLevel( logging.INFO )
  return logger
